from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import NamedTuple, Self

from jaxtyping import Array, DTypeLike, Float, Int, PRNGKeyArray

from lalamo.modules.common import LalamoModule, ParameterTree, PositionalEmbeddingSelector
from lalamo.modules.rope import PositionalEmbeddings

from .state.common import StateLayerBase

__all__ = [
    "TokenMixerBase",
    "TokenMixerConfigBase",
    "TokenMixerResult",
]


class TokenMixerResult[StateLayerT](NamedTuple):
    outputs: Float[Array, "*batch suffix_tokens channels"]
    state: StateLayerT | None = None


@dataclass(frozen=True)
class TokenMixerConfigBase(ABC):
    @property
    @abstractmethod
    def rope_dim(self) -> int | None: ...

    @abstractmethod
    def random_init(
        self,
        model_dim: int,
        *,
        key: PRNGKeyArray,
    ) -> "TokenMixerBase": ...

    @abstractmethod
    def empty(
        self,
        model_dim: int,
    ) -> "TokenMixerBase": ...


class TokenMixerBase[ConfigT, StateLayerT: StateLayerBase](LalamoModule[ConfigT]):
    @property
    @abstractmethod
    def activation_precision(self) -> DTypeLike: ...

    @property
    @abstractmethod
    def model_dim(self) -> int: ...

    @property
    @abstractmethod
    def positional_embedding_selector(self) -> PositionalEmbeddingSelector: ...

    @abstractmethod
    def __call__(
        self,
        inputs: Float[Array, "suffix_tokens channels"],
        positional_embeddings: PositionalEmbeddings | None,
        state: StateLayerT | None = None,
        return_updated_state: bool = False,
        length_without_padding: Int[Array, ""] | int | None = None,
    ) -> TokenMixerResult[StateLayerT]: ...

    @abstractmethod
    def init_static_state(self, capacity: int) -> StateLayerT: ...

    @abstractmethod
    def export_weights(self) -> ParameterTree: ...

    @abstractmethod
    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self: ...



import json
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar

import cattrs

__all__ = ["HFGenerationConfig"]


@dataclass(frozen=True)
class HFGenerationConfig:
    _converter: ClassVar[cattrs.Converter] = cattrs.Converter()
    _converter.register_structure_hook(int | list[int], lambda v, _: v)
    _converter.register_structure_hook(int | list[int] | None, lambda v, _: v)

    # -------- identity / bookkeeping --------
    _from_model_config: bool | None = None  # some Mistral & DeepSeek models
    transformers_version: str | None = None  # library version that saved the file

    # -------- special-token ids -------------
    bos_token_id: int | None = None
    eos_token_id: int | list[int] | None = None
    pad_token_id: int | None = None

    # -------- backend hints -----------------
    cache_implementation: str | None = None  # “hybrid” for Gemma 3/2

    # -------- sampling strategy -------------
    do_sample: bool | None = None
    temperature: float | None = None
    top_p: float | None = None
    top_k: int | None = None
    repetition_penalty: float | None = None

    # -------- length limits -----------------
    max_length: int | None = None  # seen in Llama 3, Gemma 2/3

    @classmethod
    def from_json(cls, json_path: Path | str) -> "HFGenerationConfig":
        json_path = Path(json_path)
        with open(json_path) as f:
            config = json.load(f)
        return cls._converter.structure(config, cls)



from lalamo.model_import.decoder_configs.huggingface import ModernBERTConfig

from .common import ConfigMap, FileSpec, ModelSpec, ModelType

__all__ = ["MIRAI_CLASSIFIER_MODELS"]

MIRAI_CLASSIFIER_MODELS = [
    ModelSpec(
        vendor="trymirai",
        family="ModernBERT",
        name="ModernBERT-Chat-Moderation",
        size="0.15B",
        quantization=None,
        repo="trymirai/chat-moderation-router",
        config_type=ModernBERTConfig,
        use_cases=tuple(),
        model_type=ModelType("classifier_model"),
        configs=ConfigMap(chat_template=FileSpec("chat_template.jinja")),
    ),
]



from .classifier import ClassifierModel, ClassifierModelConfig
from .language_model import GenerationConfig, LanguageModel, LanguageModelConfig

__all__ = [
    "ClassifierModel",
    "ClassifierModelConfig",
    "GenerationConfig",
    "LanguageModel",
    "LanguageModelConfig",
]



from .common import Speculator
from .estimator import estimate_batchsize_from_memory
from .inference import CollectTracesEvent, inference_collect_traces
from .ngram import NGramSpeculator
from .utils import SpeculatorTrainingEvent, train_speculator

__all__ = [
    "CollectTracesEvent",
    "NGramSpeculator",
    "Speculator",
    "SpeculatorTrainingEvent",
    "estimate_batchsize_from_memory",
    "inference_collect_traces",
    "train_speculator",
]



# test_registry_abc.py
import gc
import weakref

from lalamo.registry_abc import RegistryABC


def test_basic_registration_and_multilevel() -> None:
    class AbstractFoo(RegistryABC):
        pass

    class A(AbstractFoo):
        pass

    class B(A):
        pass

    # __get_descendants__ is a class method returning a list
    assert isinstance(AbstractFoo.__descendants__(), tuple)

    # Order is not guaranteed; compare as sets
    assert set(AbstractFoo.__descendants__()) >= {A, B}
    assert set(A.__descendants__()) == {B}
    assert set(B.__descendants__()) == set()


def test_excludes_classes_that_directly_list_registryabc() -> None:
    class AbstractFoo(RegistryABC):
        pass

    class A(AbstractFoo):
        pass

    # This class directly lists RegistryABC, so it must be excluded everywhere.
    class Hybrid(AbstractFoo, RegistryABC):
        pass

    assert A in AbstractFoo.__descendants__()  # tracked
    assert Hybrid not in AbstractFoo.__descendants__()  # excluded


def test_no_cross_pollination_between_roots() -> None:
    class AbstractFoo(RegistryABC):
        pass

    class AbstractBar(RegistryABC):
        pass

    class AFoo(AbstractFoo):
        pass

    class BBar(AbstractBar):
        pass

    assert AFoo in AbstractFoo.__descendants__()  # in its own tree
    assert BBar not in AbstractFoo.__descendants__()  # not leaking across

    assert BBar in AbstractBar.__descendants__()  # in its own tree
    assert AFoo not in AbstractBar.__descendants__()  # not leaking across


def test_mixin_and_indirect_registryabc_paths_are_included() -> None:
    class AbstractFoo(RegistryABC):
        pass

    class OtherRoot(RegistryABC):
        pass

    class Mixin:
        pass

    # Indirect path via a base that itself descends from RegistryABC should *not*
    # exclude the class, because the class does not directly list RegistryABC.
    class Indirect(OtherRoot):
        pass

    class D(Mixin, AbstractFoo, Indirect):
        pass

    assert D in AbstractFoo.__descendants__()


def test_dynamic_updates_on_late_definitions() -> None:
    class AbstractFoo(RegistryABC):
        pass

    assert set(AbstractFoo.__descendants__()) == set()

    class A(AbstractFoo):
        pass

    assert A in AbstractFoo.__descendants__()

    class B(A):
        pass

    # Newly defined class appears without needing to "refresh" anything
    assert B in AbstractFoo.__descendants__()
    # And it should also appear under A
    assert B in A.__descendants__()


def test_weakset_allows_garbage_collection() -> None:
    class AbstractFoo(RegistryABC):
        pass

    def make_temp() -> type:
        class Temp(AbstractFoo):
            pass

        return Temp

    temp_cls = make_temp()
    wr = weakref.ref(temp_cls)

    # It should be registered while alive
    assert temp_cls in AbstractFoo.__descendants__()
    assert wr() is not None

    # Drop the strong ref and collect
    del temp_cls
    gc.collect()

    # Weak ref is cleared, and the registry reflects that
    assert wr() is None
    names = {cls.__name__ for cls in AbstractFoo.__descendants__()}
    assert "Temp" not in names


def test_descendants_method_exists_on_all_registryabc_subclasses() -> None:
    class Root(RegistryABC):
        pass

    class Mid(Root):
        pass

    class Leaf(Mid):
        pass

    # Method exists and returns lists on all levels
    for C in (Root, Mid, Leaf):  # noqa: N806
        assert callable(C.__descendants__)
        assert isinstance(C.__descendants__(), tuple)

    assert Leaf in Mid.__descendants__()
    assert Leaf in Root.__descendants__()
    assert Mid in Root.__descendants__()



from .common import ForeignClassifierConfig, ForeignConfig, ForeignLMConfig

# from .executorch import ETLlamaConfig
from .huggingface import (
    HFGemma2Config,
    HFGemma3Config,
    HFGemma3TextConfig,
    HFGPTOssConfig,
    HFLFM2Config,
    HFLlamaConfig,
    HFLlambaConfig,
    HFMistralConfig,
    HFQwen2Config,
    HFQwen3Config,
)

__all__ = [
    "ForeignClassifierConfig",
    "ForeignConfig",
    # "ETLlamaConfig",
    "ForeignLMConfig",
    "HFGPTOssConfig",
    "HFGemma2Config",
    "HFGemma3Config",
    "HFGemma3TextConfig",
    "HFLFM2Config",
    "HFLlamaConfig",
    "HFLlambaConfig",
    "HFMistralConfig",
    "HFQwen2Config",
    "HFQwen3Config",
]



from lalamo.model_import.decoder_configs import HFLlamaConfig
from lalamo.quantization import QuantizationMode

from .common import ConfigMap, FileSpec, ModelSpec

__all__ = ["LLAMA_MODELS"]

LLAMA31 = [
    ModelSpec(
        vendor="Meta",
        family="Llama-3.1",
        name="Llama-3.1-8B-Instruct",
        size="8B",
        quantization=None,
        repo="meta-llama/Llama-3.1-8B-Instruct",
        config_type=HFLlamaConfig,
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Meta",
        family="Llama-3.1",
        name="Llama-3.1-8B-Instruct-4bit",
        size="8B",
        quantization=QuantizationMode.UINT4,
        repo="mlx-community/Llama-3.1-8B-Instruct-4bit",
        config_type=HFLlamaConfig,
        use_cases=tuple(),
        configs=ConfigMap(generation_config=FileSpec("generation_config.json", "meta-llama/Llama-3.1-8B-Instruct")),
    ),
]


LLAMA32 = [
    ModelSpec(
        vendor="Meta",
        family="Llama-3.2",
        name="Llama-3.2-1B-Instruct",
        size="1B",
        quantization=None,
        repo="meta-llama/Llama-3.2-1B-Instruct",
        config_type=HFLlamaConfig,
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Meta",
        family="Llama-3.2",
        name="Llama-3.2-1B-Instruct-4bit",
        size="1B",
        quantization=QuantizationMode.UINT4,
        repo="mlx-community/Llama-3.2-1B-Instruct-4bit",
        config_type=HFLlamaConfig,
        configs=ConfigMap(generation_config=FileSpec("generation_config.json", "meta-llama/Llama-3.2-1B-Instruct")),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Meta",
        family="Llama-3.2",
        name="Llama-3.2-1B-Instruct-8bit",
        size="1B",
        quantization=QuantizationMode.UINT8,
        repo="mlx-community/Llama-3.2-1B-Instruct-8bit",
        config_type=HFLlamaConfig,
        configs=ConfigMap(generation_config=FileSpec("generation_config.json", "meta-llama/Llama-3.2-1B-Instruct")),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Meta",
        family="Llama-3.2",
        name="Llama-3.2-3B-Instruct",
        size="3B",
        quantization=None,
        repo="meta-llama/Llama-3.2-3B-Instruct",
        config_type=HFLlamaConfig,
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Meta",
        family="Llama-3.2",
        name="Llama-3.2-3B-Instruct-4bit",
        size="3B",
        quantization=QuantizationMode.UINT4,
        repo="mlx-community/Llama-3.2-3B-Instruct-4bit",
        config_type=HFLlamaConfig,
        configs=ConfigMap(generation_config=FileSpec("generation_config.json", "meta-llama/Llama-3.2-3B-Instruct")),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Meta",
        family="Llama-3.2",
        name="Llama-3.2-3B-Instruct-8bit",
        size="3B",
        quantization=QuantizationMode.UINT8,
        repo="mlx-community/Llama-3.2-3B-Instruct-8bit",
        config_type=HFLlamaConfig,
        configs=ConfigMap(generation_config=FileSpec("generation_config.json", "meta-llama/Llama-3.2-3B-Instruct")),
        use_cases=tuple(),
    ),
]

LLAMA_MODELS = LLAMA31 + LLAMA32



from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from typing import NamedTuple

import equinox as eqx
import jax
import jax.numpy as jnp
from einops import rearrange
from jax import vmap
from jaxtyping import Array, Bool, Float, Int, PRNGKeyArray

from lalamo.message_processor import AssistantMessage, Message, MessageProcessor
from lalamo.modules import (
    Decoder,
    DecoderConfig,
    DecoderForwardPassConfig,
    ForwardPassMode,
    LalamoModule,
    State,
)
from lalamo.sampling import SamplingPolicy, make_policy

from .common import TextModel, TextModelConfig

__all__ = [
    "ForwardPassConfig",
    "GenerationConfig",
    "LanguageModel",
    "LanguageModelConfig",
]


_COMPILED_PROMPT_LENGTHS = [512 * 2**i for i in range(10)]


type ForwardPassConfig = DecoderForwardPassConfig


class PrefillResults(NamedTuple):
    last_token_logits: Float[Array, "batch vocabulary"]
    last_token_indices: Int[Array, " batch"]
    state: State


class DecodingState(NamedTuple):
    last_token_logits: Float[Array, "batch vocabulary"]
    last_token_indices: Int[Array, " batch"]
    state: State
    stop_flags: Bool[Array, " batch"]


class GenerationStepResults(NamedTuple):
    token_ids: Int[Array, " batch"]
    top_k_token_ids: Int[Array, " batch k"] | None
    top_k_token_logits: Float[Array, " batch k"] | None


class GenerationResults(NamedTuple):
    token_ids: Int[Array, "batch response_tokens"]
    top_k_token_ids: Int[Array, "batch response_tokens k"] | None
    top_k_token_logits: Float[Array, "batch response_tokens k"] | None


@dataclass(frozen=True)
class GenerationConfig:
    stop_token_ids: tuple[int, ...]
    temperature: float | None
    top_k: int | None
    top_p: float | None
    banned_tokens: tuple[int, ...] | None

    def default_policy(self) -> SamplingPolicy:
        return make_policy(self.temperature, self.top_k, self.top_p, self.banned_tokens)


@dataclass(frozen=True)
class LanguageModelConfig(TextModelConfig[DecoderConfig]):
    generation_config: GenerationConfig

    def init(
        self,
        model: LalamoModule,
        message_processor: MessageProcessor,
    ) -> "LanguageModel":
        assert isinstance(model, Decoder)
        return LanguageModel(self, model, message_processor)

    @classmethod
    def load_model(cls, path: Path | str) -> "LanguageModel":
        result = super().load_model(path)
        assert isinstance(result, LanguageModel)
        return result


class LanguageModel(TextModel[LanguageModelConfig, Decoder]):
    @property
    def stop_token_ids(self) -> tuple[int, ...]:
        return self.config.generation_config.stop_token_ids

    def default_sampling_policy(self) -> SamplingPolicy:
        return self.config.generation_config.default_policy()

    @eqx.filter_jit
    def _prefill(
        self,
        token_ids: Int[Array, "batch tokens"],
        lengths_without_padding: Int[Array, " batch"] | None = None,
        state_capacity: int | None = None,
        forward_pass_config: ForwardPassConfig | None = None,
    ) -> PrefillResults:
        batch_size, sequence_length = token_ids.shape
        token_positions = jnp.repeat(jnp.arange(sequence_length, dtype=jnp.int32)[None, ...], batch_size, axis=0)
        if state_capacity is not None:
            state = self.model.init_static_state(batch_size, state_capacity)
        else:
            state = None

        decoder_outputs = self.model(
            token_ids,
            token_positions,
            state,
            return_updated_state=True,
            lengths_without_padding=lengths_without_padding,
            forward_pass_mode=ForwardPassMode.MULTI_TOKEN,
            forward_pass_config=forward_pass_config,
        )

        if lengths_without_padding is not None:
            last_logits_indices = lengths_without_padding - 1
        else:
            last_logits_indices = jnp.array([sequence_length - 1] * batch_size, dtype=jnp.int32)

        last_token_logits = vmap(lambda logits, index: logits[index])(decoder_outputs.logits, last_logits_indices)

        assert decoder_outputs.updated_state is not None
        return PrefillResults(
            last_token_logits=last_token_logits,
            last_token_indices=last_logits_indices,
            state=decoder_outputs.updated_state,
        )

    @eqx.filter_jit
    def generate_tokens(
        self,
        prompt_token_ids: Int[Array, "batch prompt_tokens"],
        sampling_policy: SamplingPolicy | None = None,
        prompt_lengths_without_padding: Int[Array, " batch"] | None = None,
        max_output_length: int = 8192,
        eos_token_ids: Int[Array, " eos_tokens"] | None = None,
        forward_pass_config: ForwardPassConfig | None = None,
        num_top_logits_to_return: int | None = None,
        *,
        key: PRNGKeyArray | None = None,
    ) -> GenerationResults:
        if sampling_policy is None:
            sampling_policy = self.default_sampling_policy()
        if eos_token_ids is None:
            eos_token_ids = jnp.array(self.stop_token_ids, dtype=jnp.int32)

        batch_size, sequence_length = prompt_token_ids.shape
        prefill_results = self._prefill(
            prompt_token_ids,
            prompt_lengths_without_padding,
            sequence_length + max_output_length,
            forward_pass_config=forward_pass_config,
        )

        initial_state = DecodingState(
            prefill_results.last_token_logits,
            prefill_results.last_token_indices,
            prefill_results.state,
            jnp.zeros(batch_size, dtype=jnp.bool),
        )

        if key is None:
            key = jax.random.PRNGKey(0)
        keys = jax.random.split(key, num=max_output_length)

        def loop_iteration(
            state: DecodingState,
            key: PRNGKeyArray,
        ) -> tuple[DecodingState, GenerationStepResults]:
            def sample_and_update() -> tuple[DecodingState, GenerationStepResults]:
                upcasted_logits = state.last_token_logits.astype(jnp.float32)
                processed_logits = vmap(sampling_policy.process_logits)(upcasted_logits)
                next_token_ids = jax.random.categorical(key, processed_logits)
                next_token_ids = jnp.where(state.stop_flags, jnp.zeros(batch_size, dtype=jnp.int32), next_token_ids)
                if num_top_logits_to_return is not None:
                    next_top_k_token_logits, next_top_k_token_ids = jax.lax.top_k(
                        processed_logits,
                        num_top_logits_to_return,
                    )
                else:
                    next_top_k_token_ids = None
                    next_top_k_token_logits = None
                next_token_indices = state.last_token_indices + 1

                stop_flags = state.stop_flags | jnp.any(next_token_ids[:, None] == eos_token_ids[None, :], axis=-1)

                if batch_size == 1:
                    forward_pass_mode = ForwardPassMode.SINGLE_TOKEN
                else:
                    forward_pass_mode = ForwardPassMode.MULTI_TOKEN

                decoder_outputs = self.model(
                    next_token_ids[:, None],
                    next_token_indices[:, None],
                    state.state,
                    return_updated_state=True,
                    forward_pass_mode=forward_pass_mode,
                    forward_pass_config=forward_pass_config,
                )
                assert decoder_outputs.updated_state is not None, "updated_state should not be None"
                new_state = DecodingState(
                    decoder_outputs.logits.squeeze(1),
                    next_token_indices,
                    decoder_outputs.updated_state,
                    stop_flags,
                )
                return new_state, GenerationStepResults(next_token_ids, next_top_k_token_ids, next_top_k_token_logits)

            def pad_and_repeat_state() -> tuple[DecodingState, GenerationStepResults]:
                (batch_size,) = state.stop_flags.shape
                pad_token = jnp.zeros(batch_size, dtype=jnp.int32)
                if num_top_logits_to_return is not None:
                    top_k_token_ids = jnp.zeros((batch_size, num_top_logits_to_return), dtype=jnp.int32)
                    top_k_token_logits = jnp.zeros((batch_size, num_top_logits_to_return), dtype=jnp.float32)
                else:
                    top_k_token_ids = None
                    top_k_token_logits = None
                return state, GenerationStepResults(pad_token, top_k_token_ids, top_k_token_logits)

            return jax.lax.cond(jnp.all(state.stop_flags), pad_and_repeat_state, sample_and_update)

        _, generated = jax.lax.scan(loop_iteration, initial_state, keys)

        token_ids = rearrange(generated.token_ids, "iteration batch -> batch iteration")

        if num_top_logits_to_return is not None:
            top_k_token_ids = rearrange(generated.top_k_token_ids, "iteration batch k -> batch iteration k")
            top_k_token_logits = rearrange(generated.top_k_token_logits, "iteration batch k -> batch iteration k")
        else:
            top_k_token_ids = None
            top_k_token_logits = None

        return GenerationResults(token_ids, top_k_token_ids, top_k_token_logits)

    def reply(
        self,
        messages: Iterable[Message],
        sampling_policy: SamplingPolicy | None = None,
        forward_pass_config: ForwardPassConfig | None = None,
        *,
        key: PRNGKeyArray | None = None,
    ) -> AssistantMessage:
        formatted_messages = self.message_processor.render_request(messages)
        token_ids = jnp.array(self.message_processor.tokenize_text(formatted_messages), dtype=jnp.int32)[None, :]
        response_ids = self.generate_tokens(
            token_ids,
            sampling_policy,
            forward_pass_config=forward_pass_config,
            key=key,
        ).token_ids.squeeze(0)
        response_text = self.message_processor.detokenize(response_ids.tolist())
        return self.message_processor.parse_response(response_text)

    def stream_reply_text(
        self,
        messages: Iterable[Message],
        sampling_policy: SamplingPolicy | None = None,
        max_output_length: int = 8192,
        forward_pass_config: ForwardPassConfig | None = None,
        *,
        key: PRNGKeyArray | None = None,
    ) -> Iterable[str]:
        formatted_messages = self.message_processor.render_request(messages)
        token_ids = jnp.array(self.message_processor.tokenize_text(formatted_messages), dtype=jnp.int32)
        for token_id in self.stream_tokens(
            token_ids,
            sampling_policy,
            max_output_length,
            forward_pass_config=forward_pass_config,
            key=key,
        ):
            yield self.message_processor.detokenize([token_id.item()])

    def stream_tokens(
        self,
        prompt_token_ids: Int[Array, " prompt_tokens"],
        sampling_policy: SamplingPolicy | None = None,
        max_output_length: int = 8192,
        eos_token_ids: Int[Array, " eos_tokens"] | None = None,
        forward_pass_config: ForwardPassConfig | None = None,
        *,
        key: PRNGKeyArray | None = None,
    ) -> Iterable[Int[Array, ""]]:
        if sampling_policy is None:
            sampling_policy = self.default_sampling_policy()
        if eos_token_ids is None:
            eos_token_ids = jnp.array(self.stop_token_ids, dtype=jnp.int32)

        (input_length,) = prompt_token_ids.shape

        padded_input_length = min(length for length in _COMPILED_PROMPT_LENGTHS if length >= input_length)
        padded_token_ids = jnp.zeros((padded_input_length,), dtype=jnp.int32)
        padded_token_ids = padded_token_ids.at[:input_length].set(prompt_token_ids)

        prefill_results = self._prefill(
            padded_token_ids[None, :],
            jnp.array([input_length], dtype=jnp.int32),
            padded_input_length + max_output_length,
            forward_pass_config=forward_pass_config,
        )

        if key is None:
            key = jax.random.PRNGKey(0)
        keys = jax.random.split(key, num=max_output_length)

        state = DecodingState(
            prefill_results.last_token_logits,
            prefill_results.last_token_indices,
            prefill_results.state,
            jnp.array([0], dtype=jnp.bool),
        )

        for iter_key in keys:
            upcasted_logits = state.last_token_logits.astype(jnp.float32)
            processed_logits = sampling_policy.process_logits(upcasted_logits.squeeze(0))
            next_token_id = jax.random.categorical(iter_key, processed_logits)

            yield next_token_id

            if jnp.any(next_token_id == eos_token_ids):
                return

            next_token_indices = state.last_token_indices + 1
            decoder_outputs = self.model(
                next_token_id.reshape(1, 1),
                next_token_indices.reshape(1, 1),
                state.state,
                return_updated_state=True,
                forward_pass_config=forward_pass_config,
            )
            assert decoder_outputs.updated_state is not None, "updated_state should not be None"
            state = DecodingState(
                decoder_outputs.logits.squeeze(1),
                next_token_indices,
                decoder_outputs.updated_state,
                state.stop_flags,
            )



from lalamo.model_import.decoder_configs.huggingface import HFGemma3TextConfig

from .common import ModelSpec

__all__ = ["RNJ_MODELS"]

RNJ_MODELS = [
    ModelSpec(
        vendor="EssentialAI",
        family="Rnj-1",
        name="Rnj-1-Instruct",
        size="8B",
        quantization=None,
        repo="EssentialAI/rnj-1-instruct",
        config_type=HFGemma3TextConfig,
    ),
]



import jax
import jax.numpy as jnp

from lalamo.modules import (
    GELU,
    DenseMLPConfig,
    ForwardPassMode,
    FullPrecisionLinearConfig,
    MixtureOfExpertsConfig,
    SoftmaxRouting,
)

from .common import assert_close


def test_moe_prefill_vs_decode_match() -> None:
    # Model hyperparameters (kept small for speed)
    model_dim = 64
    hidden_dim = 512
    mixture_size = 16
    num_experts_per_token = 2

    # Build Mixture-of-Experts config with full-precision linear layers
    router_config = FullPrecisionLinearConfig(precision=jnp.float32)
    expert_linear_config = FullPrecisionLinearConfig(precision=jnp.float32)
    expert_config = DenseMLPConfig(
        linear_config=expert_linear_config,
        activation=GELU(),
        has_up_biases=False,
        has_down_biases=False,
        gate_clipping=None,
        up_clipping=None,
    )

    moe_config = MixtureOfExpertsConfig(
        mixture_size=mixture_size,
        num_experts_per_token=num_experts_per_token,
        routing_function=SoftmaxRouting(),
        router_config=router_config,
        router_has_biases=False,
        expert_config=expert_config,
    )

    # Initialize MoE
    key = jax.random.PRNGKey(0)
    moe = moe_config.random_init(model_dim=model_dim, hidden_dim=hidden_dim, key=key)

    # Random inputs
    batch = 2
    suffix_tokens = 16
    inputs_key = jax.random.PRNGKey(1)
    inputs = jax.random.normal(inputs_key, (batch, suffix_tokens, model_dim), dtype=jnp.float32)

    # Compare PREFILL vs DECODE
    out_prefill = moe(inputs, forward_pass_mode=ForwardPassMode.MULTI_TOKEN)
    out_decode = moe(inputs, forward_pass_mode=ForwardPassMode.SINGLE_TOKEN)

    assert_close(result=out_decode, reference=out_prefill)



import functools
from collections.abc import Callable, Iterable
from itertools import batched, chain
from typing import NamedTuple

import jax
import jax.numpy as jnp

from lalamo.data.lalamo_completions import LalamoCompletion
from lalamo.data.utils import get_prefixes_ending_in_user_message
from lalamo.message_processor import Message
from lalamo.models import LanguageModel


class CollectTracesEvent(NamedTuple):
    sequences_processed: int
    tokens_generated: int


def inference_collect_traces(
    model: LanguageModel,
    conversations: Iterable[Iterable[Message]],
    num_top_logits_to_collect: int = 8,
    batch_size: int = 1,
    max_input_length: int = 1024,
    max_output_length: int = 1024,
    tokens_to_generate: int | None = None,
    progress_callback: Callable[[CollectTracesEvent], None] | None = None,
) -> Iterable[LalamoCompletion]:
    generate_tokens_compiled = (
        jax.jit(
            functools.partial(
                LanguageModel.generate_tokens,
                max_output_length=max_output_length,
                num_top_logits_to_return=num_top_logits_to_collect,
            ),
        )
        .lower(
            model,
            prompt_token_ids=jax.ShapeDtypeStruct((batch_size, max_input_length), jnp.int32),
            prompt_lengths_without_padding=jax.ShapeDtypeStruct((batch_size,), jnp.int32),
        )
        .compile()
    )

    prefixes = chain.from_iterable(map(get_prefixes_ending_in_user_message, conversations))

    tokenized_prefixes = map(model.message_processor.tokenize_request, prefixes)
    filtered_prefixes = filter(lambda conv: len(conv) <= max_input_length, tokenized_prefixes)

    tokens_generated, sequences_processed = 0, 0

    for real_batch in batched(filtered_prefixes, n=batch_size):
        batch_padding = batch_size - len(real_batch)
        batch = (*real_batch, *(([0],) * batch_padding))

        length_without_padding = jnp.array(list(map(len, batch)))

        padded = jnp.array(
            [jnp.pad(jnp.array(tokens), (0, max_input_length - len(tokens)), constant_values=0) for tokens in batch],
        )

        generated = generate_tokens_compiled(
            model,
            prompt_token_ids=padded,
            prompt_lengths_without_padding=length_without_padding,
        )

        assert generated.top_k_token_ids is not None and generated.top_k_token_logits is not None

        for conv_idx in range(len(real_batch)):
            token_ids = generated.token_ids[conv_idx].tolist()
            seqlen = next((i + 1 for i, t in enumerate(token_ids) if t in model.stop_token_ids), len(token_ids))
            if tokens_to_generate is not None:
                seqlen = min(seqlen, tokens_to_generate - tokens_generated)
            tokens_generated += seqlen
            sequences_processed += 1

            token_ids = token_ids[:seqlen]
            token_logits_ids = generated.top_k_token_ids[conv_idx, : len(token_ids)].tolist()
            token_logits_values = generated.top_k_token_logits[conv_idx, : len(token_ids)].tolist()
            token_logits = [
                dict(zip(keys, values, strict=True))
                for keys, values in zip(token_logits_ids, token_logits_values, strict=True)
            ]

            yield LalamoCompletion(batch[conv_idx], token_ids, token_logits)

            if tokens_to_generate is not None and tokens_generated >= tokens_to_generate:
                break

        if progress_callback is not None:
            progress_callback(CollectTracesEvent(sequences_processed, tokens_generated))

        if tokens_to_generate is not None and tokens_generated >= tokens_to_generate:
            break



from collections.abc import Iterable
from dataclasses import dataclass
from typing import IO, Any, ClassVar, Self

import msgpack
from cattrs.preconf.msgpack import MsgpackConverter
from cattrs.preconf.msgpack import make_converter as make_msgpack_converter


@dataclass(frozen=True)
class LalamoCompletion:
    _converter: ClassVar[MsgpackConverter] = make_msgpack_converter()

    prefix_token_ids: list[int]
    completion_token_ids: list[int]
    completion_token_logits: list[dict[int, float]]

    def __post_init__(self) -> None:
        if len(self.completion_token_ids) != len(self.completion_token_logits):
            raise ValueError(f"({len(self.completion_token_ids)=}) != ({len(self.completion_token_logits)=})")

    def serialize(self) -> bytes:
        return self._converter.dumps(self)

    @classmethod
    def deserialize(cls, data: bytes | IO[bytes]) -> Self:
        if isinstance(data, bytes):
            obj: Any = msgpack.unpackb(data, strict_map_key=False)
        else:
            obj = msgpack.unpack(data, strict_map_key=False)

        return cls._converter.structure(obj, cls)

    @classmethod
    def deserialize_many(cls, data: bytes | IO[bytes]) -> Iterable[Self]:
        if isinstance(data, bytes):
            unpacker = msgpack.Unpacker(strict_map_key=False)
            unpacker.feed(data)
        else:
            unpacker = msgpack.Unpacker(file_like=data, strict_map_key=False)

        for obj in unpacker:
            yield cls._converter.structure(obj, cls)



from abc import ABC, ABCMeta
from typing import Any
from weakref import WeakSet

__all__ = ["RegistryABC", "RegistryMeta"]


class RegistryMeta(ABCMeta):
    """
    Metaclass that tracks, for each subclass of RegistryABC, a per-class WeakSet
    of descendants (classes that have it in their MRO) while excluding any class
    that directly lists `RegistryABC` among its bases.
    """

    _REG_ATTR: str = "__registry_descendants__"
    _ROOT: type["RegistryABC"] | None = None

    def __init__(
        cls: type,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, object],
        **kwargs: Any,  # noqa: ANN401
    ) -> None:
        super().__init__(name, bases, namespace, **kwargs)  # type: ignore[call-overload]

        # Give *every* class its own WeakSet (shadow any inherited attribute)
        setattr(cls, RegistryMeta._REG_ATTR, WeakSet())

        # Detect and remember the root exactly once
        if RegistryMeta._ROOT is None and name == "RegistryABC":
            RegistryMeta._ROOT = cls  # type: ignore[assignment]
            return

        root = RegistryMeta._ROOT
        if root is None:
            # Extremely early import edge-case; nothing to register yet
            return

        # Exclude classes that directly list the root among bases
        if any(b is root for b in cls.__bases__):
            return

        # Register this class on all qualifying ancestors below the root
        for ancestor in cls.__mro__[1:]:
            if isinstance(ancestor, RegistryMeta) and issubclass(ancestor, root):
                getattr(ancestor, RegistryMeta._REG_ATTR).add(cls)


class RegistryABC(ABC, metaclass=RegistryMeta):
    """
    Abstract base tracked by RegistryMeta.

    Any class defined as `class AbstractFoo(RegistryABC)` will expose a
    class method `AbstractFoo.__get_descendants__()` that returns a list of
    all classes having AbstractFoo in their MRO *except* those that directly
    include `RegistryABC` among their bases.
    """

    @classmethod
    def __descendants__(cls) -> tuple[type, ...]:
        reg: WeakSet[type] = getattr(cls, RegistryMeta._REG_ATTR)  # noqa: SLF001
        return tuple(reg)



from collections.abc import Mapping
from dataclasses import dataclass

import jax.numpy as jnp
from einops import rearrange
from jaxtyping import Array, DTypeLike

from lalamo.common import ParameterPath
from lalamo.modules import (
    Attention,
    AttentionConfig,
    Decoder,
    DenseMLP,
    FullPrecisionLinear,
    GroupQuantizedLinear,
    LinearBase,
    Mamba2,
    Mamba2Config,
    MLXQuantizedLinear,
    MLXQuantizedTiedEmbedding,
    MLXQuantizedTiedEmbeddingConfig,
    MLXSemiQuantizedUntiedEmbedding,
    Normalization,
    SeparableCausalConv,
    ShortConv,
    ShortConvConfig,
    TiedEmbedding,
    TransformerLayer,
    UntiedEmbedding,
)
from lalamo.modules.classifier import Classifier
from lalamo.modules.embedding import MLXQuantizedUntiedEmbedding
from lalamo.modules.mlp import MixtureOfExperts, MLPBase
from lalamo.quantization import QuantizationMode

from .common import load_parameters
from .utils import decode_mxfp4, deinterleave_pairwise_columns

__all__ = ["load_huggingface_decoder"]


AWQ_UINT4_REVERSE_ORDER = jnp.array([0, 4, 1, 5, 2, 6, 3, 7], dtype=jnp.int32)


def _reverse_uint4_order(array: Array, reverse_order: Array) -> Array:
    """Reverses the AWQ packing order to get the logical order of channels for INT4."""
    pack_factor = 32 // 4
    *_, last_dim = array.shape
    if last_dim % pack_factor != 0:
        return array

    array_reshaped = rearrange(
        array,
        "... (group pack_factor) -> ... group pack_factor",
        pack_factor=pack_factor,
    )
    array_reordered = array_reshaped[..., reverse_order]
    return rearrange(array_reordered, "... group pack_factor -> ... (group pack_factor)")


def unpack_int32(packed_weights: Array, mode: QuantizationMode) -> Array:
    assert packed_weights.dtype in (
        jnp.int32,
        jnp.uint32,
    ), f"Expected packed_weights to be of dtype jnp.(u)int32, got {packed_weights.dtype}"
    assert 32 % mode.bits == 0

    shifts = jnp.arange(0, 32, mode.bits)
    mask = (2**mode.bits) - 1
    unpacked = jnp.bitwise_and(jnp.right_shift(packed_weights[:, :, None], shifts[None, None, :]), mask)
    unpacked = rearrange(
        unpacked,
        "out_channels packed_groups packed_values -> out_channels (packed_groups packed_values)",
    )

    return unpacked


def _process_quantized_tensor(
    quantized: Array,
    weight_quantization: QuantizationMode,
    activation_precision: DTypeLike,
    reverse_order: Array | None = None,
) -> Array:
    unpacked = unpack_int32(quantized, weight_quantization)
    if reverse_order is not None:
        assert weight_quantization == QuantizationMode.UINT4, "reverse order only supported on uint4 quant type"
        unpacked = _reverse_uint4_order(unpacked, reverse_order)

    return unpacked.astype(activation_precision)


def _fuse_full_precision_weights(
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
    sublayers_to_fuse: list[str] | None,
) -> Array:
    if sublayers_to_fuse is None:
        return weights_dict[path / "weight"]

    weights = [weights_dict[path / layer_name / "weight"] for layer_name in sublayers_to_fuse]
    return jnp.concatenate(weights, axis=0)


@dataclass(frozen=True)
class QuantizedParamLayout:
    weight: str
    scale: str
    bias: str
    transposed: bool


AWQ_QUANTIZED_WEIGHT_LAYOUT = QuantizedParamLayout("qweight", "scales", "qzeros", transposed=True)
MLX_QUANTIZED_WEIGHT_LAYOUT = QuantizedParamLayout("weight", "scales", "biases", transposed=False)


def _fuse_quantized_weights(
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
    sublayers_to_fuse: list[str] | None,
    quantized_param_layout: QuantizedParamLayout,
) -> tuple[Array, Array, Array]:
    # Note that AWQ quantized weights are stored transposed relative to full-precision weights

    if sublayers_to_fuse is None:
        qweights = weights_dict[path / quantized_param_layout.weight]
        qzeros = weights_dict[path / quantized_param_layout.bias]
        scales = weights_dict[path / quantized_param_layout.scale]
        return qweights, qzeros, scales

    qweights = [weights_dict[path / layer_name / quantized_param_layout.weight] for layer_name in sublayers_to_fuse]
    qzeros = [weights_dict[path / layer_name / quantized_param_layout.bias] for layer_name in sublayers_to_fuse]
    scales = [weights_dict[path / layer_name / quantized_param_layout.scale] for layer_name in sublayers_to_fuse]

    fused_qweights = jnp.concatenate(qweights, axis=int(quantized_param_layout.transposed))
    fused_qzeros = jnp.concatenate(qzeros, axis=int(quantized_param_layout.transposed))
    fused_scales = jnp.concatenate(scales, axis=int(quantized_param_layout.transposed))

    return fused_qweights, fused_qzeros, fused_scales


def load_linear(
    module: LinearBase,
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
    sublayers_to_fuse: list[str] | None = None,
) -> LinearBase:
    """Loads a linear layer, optionally fusing weights from sublayers."""
    if not module.has_biases:
        if sublayers_to_fuse:
            paths_to_check = [path / proj / "bias" for proj in sublayers_to_fuse]
        else:
            paths_to_check = path / "bias"
        for p in paths_to_check:
            if p in weights_dict:
                raise ValueError(f"Bias tensor found at {p} but module does not support it.")
        bias = None
    elif sublayers_to_fuse is None:
        bias = weights_dict[path / "bias"]
    else:
        bias = jnp.concatenate(
            [weights_dict[path / proj_name / "bias"] for proj_name in sublayers_to_fuse],
            axis=0,
        )

    if isinstance(module, FullPrecisionLinear):
        weights = _fuse_full_precision_weights(weights_dict, path, sublayers_to_fuse)
        return load_parameters(lambda m: (m.weights, m.biases), module, (weights, bias))

    if isinstance(module, GroupQuantizedLinear):
        qweights, qzeros, scales = _fuse_quantized_weights(
            weights_dict,
            path,
            sublayers_to_fuse,
            AWQ_QUANTIZED_WEIGHT_LAYOUT,
        )
        weight_quantization = module.config.weight_quantization_mode
        activation_precision = module.activation_precision

        if weight_quantization == QuantizationMode.UINT4:
            reverse_order = AWQ_UINT4_REVERSE_ORDER
        else:
            reverse_order = None

        weights = _process_quantized_tensor(
            qweights,
            weight_quantization,
            activation_precision,
            reverse_order,
        )
        zeros = _process_quantized_tensor(
            qzeros,
            weight_quantization,
            activation_precision,
            reverse_order,
        )
        scales = scales.astype(activation_precision)

        return load_parameters(
            lambda m: (m.weights, m.scales, m.zero_points, m.biases),
            module,
            (weights.T, scales.T, zeros.T, bias),
        )

    if isinstance(module, MLXQuantizedLinear):
        qweights, deq_biases, scales = _fuse_quantized_weights(
            weights_dict,
            path,
            sublayers_to_fuse,
            MLX_QUANTIZED_WEIGHT_LAYOUT,
        )
        weight_quantization = module.config.weight_quantization_mode
        activation_precision = module.activation_precision

        weights = _process_quantized_tensor(
            qweights,
            weight_quantization,
            activation_precision,
            None,
        )
        scales = scales.astype(activation_precision)
        deq_biases = deq_biases.astype(activation_precision)

        return load_parameters(
            lambda m: (m.weights, m.scales, m.deq_biases, m.biases),
            module,
            (weights, scales, deq_biases, bias),
        )

    raise TypeError(f"Unsupported module type for loading: {type(module)}")


def load_mlp(
    module: MLPBase,
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
    up_proj_key: str,
    gate_proj_key: str,
    down_proj_key: str,
) -> MLPBase:
    if isinstance(module, DenseMLP):
        # Standard dense MLP with separate sublayers.
        up_projection = load_linear(
            module.up_projection,
            weights_dict,
            path,
            sublayers_to_fuse=[up_proj_key, gate_proj_key],
        )
        down_projection = load_linear(module.down_projection, weights_dict, path / down_proj_key)
        return load_parameters(
            lambda m: (m.up_projection, m.down_projection),
            module,
            (up_projection, down_projection),
        )

    if isinstance(module, MixtureOfExperts):
        return load_moe(module, weights_dict, path)

    raise TypeError(f"Unsupported module type for loading: {type(module)}")


def load_moe(module: MixtureOfExperts, weights_dict: Mapping[str, Array], path: ParameterPath) -> MixtureOfExperts:
    # Load router via the standard linear loader
    router = load_linear(module.router, weights_dict, path / "router")

    experts_path = path / "experts"
    # Handle fused MXFP4 experts layout if present
    if (experts_path / "gate_up_proj_blocks") in weights_dict:
        # Decode fused gate/up (interleaved), split into (up, gate), and add +1.0 to up bias
        fused = decode_mxfp4(
            weights_dict[experts_path / "gate_up_proj_blocks"],
            weights_dict[experts_path / "gate_up_proj_scales"],
            dtype=module.activation_precision,
            flatten=False,
        )
        # Stored as (experts, outputs=2*hidden_dim, input_blocks, input_block_elems)
        # Merge blocks and move outputs last
        fused_eio = rearrange(fused, "e o ib ie -> e (ib ie) o")
        up_w, gate_w = deinterleave_pairwise_columns(fused_eio, first="odd")
        combined_up_gate = jnp.concatenate([up_w, gate_w], axis=-1)
        # Transpose to new layout: (experts, outputs, inputs)
        combined_up_gate_w = jnp.swapaxes(combined_up_gate, -1, -2)

        gub = weights_dict[experts_path / "gate_up_proj_bias"]
        if gub.ndim == 1:
            # Broadcast to (experts, 2*hidden_dim)
            gub = jnp.broadcast_to(gub, (combined_up_gate_w.shape[0], gub.shape[0]))
        up_b, gate_b = deinterleave_pairwise_columns(gub, first="odd")
        combined_up_gate_b = jnp.concatenate([up_b + 1.0, gate_b], axis=-1)

        up_projection = load_parameters(
            lambda m: (m.weights, m.biases),  # type: ignore
            module.experts.up_projection,
            (combined_up_gate_w, combined_up_gate_b),
        )

        # Down projection: decode MXFP4 to dense
        down_w = decode_mxfp4(
            weights_dict[experts_path / "down_proj_blocks"],
            weights_dict[experts_path / "down_proj_scales"],
            dtype=module.activation_precision,
            flatten=False,
        )
        # Stored as (experts, outputs=model_dim, input_blocks, input_block_elems)
        # Merge blocks and move outputs last
        down_w = rearrange(down_w, "e o ib ie -> e o (ib ie)")
        down_b = weights_dict[experts_path / "down_proj_bias"]
        if down_b.ndim == 1:
            down_b = jnp.broadcast_to(down_b, (*down_w.shape[:-1], down_b.shape[0]))

        down_projection = load_parameters(
            lambda m: (m.weights, m.biases),  # type: ignore
            module.experts.down_projection,
            (down_w, down_b),
        )

        experts = load_parameters(
            lambda m: (m.up_projection, m.down_projection),
            module.experts,
            (up_projection, down_projection),
        )
    else:
        # Fallback: recursively load a standard DenseMLP experts module
        experts = load_mlp(
            module.experts,
            weights_dict,
            experts_path,
            "up_proj",
            "gate_proj",
            "down_proj",
        )

    return load_parameters(
        lambda m: (m.router, m.experts),
        module,
        (router, experts),
    )


def load_rmsnorm(
    module: Normalization,
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
) -> Normalization:
    scales = weights_dict[path / "weight"]
    return load_parameters(lambda m: (m.scales,), module, (scales,))


def load_attention(
    module: Attention,
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
) -> Attention:
    if (path / "o_proj.weight") in weights_dict or (path / "o_proj.qweight") in weights_dict:
        o_proj_name = "o_proj"
    elif (path / "out_proj.weight") in weights_dict or (path / "out_proj.qweight") in weights_dict:
        o_proj_name = "out_proj"
    else:
        raise NotImplementedError("Can't determine attention output projection name")

    qkv_projection = load_linear(
        module.qkv_projection,
        weights_dict,
        path,
        sublayers_to_fuse=["q_proj", "k_proj", "v_proj"],
    )
    out_projection = load_linear(module.out_projection, weights_dict, path / o_proj_name)

    if module.query_norm is not None:
        if (path / "q_norm.weight") in weights_dict:
            q_norm_name = "q_norm"
        elif (path / "q_layernorm.weight") in weights_dict:
            q_norm_name = "q_layernorm"
        else:
            raise NotImplementedError("Can't determine attention query projection parameter name")

        query_norm = load_rmsnorm(module.query_norm, weights_dict, path / q_norm_name)
    else:
        query_norm = None

    if module.key_norm is not None:
        if (path / "k_norm.weight") in weights_dict:
            k_norm_name = "k_norm"
        elif (path / "k_layernorm.weight") in weights_dict:
            k_norm_name = "k_layernorm"
        else:
            raise NotImplementedError("Can't determine attention key projection parameter name")

        key_norm = load_rmsnorm(module.key_norm, weights_dict, path / k_norm_name)
    else:
        key_norm = None

    # GPT-OSS adds per-head attention sinks; load them if present.
    if (path / "sinks") in weights_dict:
        sinks = weights_dict[path / "sinks"]
    else:
        sinks = module.sinks

    return load_parameters(
        lambda m: (
            m.qkv_projection,
            m.out_projection,
            m.query_norm,
            m.key_norm,
            m.sinks,
        ),
        module,
        (qkv_projection, out_projection, query_norm, key_norm, sinks),
    )


def _load_conv(
    conv_module: SeparableCausalConv,
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
    permute_conv: bool,
) -> SeparableCausalConv:
    weight_path = path / "conv1d" / "weight"
    if weight_path not in weights_dict:
        weight_path = path / "conv_weight"
    if weight_path not in weights_dict:
        weight_path = path / "conv.weight"
    if weight_path not in weights_dict:
        weight_path = None

    if weight_path is not None:
        raw = weights_dict[weight_path]
        if permute_conv:
            raw = jnp.matrix_transpose(raw)
        conv_weight = raw.squeeze(1) if raw.ndim == 3 else raw
    else:
        conv_weight = conv_module.weights

    bias_path = path / "conv1d" / "bias"
    if bias_path not in weights_dict:
        bias_path = path / "conv_bias"
    if bias_path not in weights_dict:
        bias_path = path / "conv.bias"
    if bias_path not in weights_dict:
        bias_path = None

    if bias_path is not None and conv_module.biases is not None:
        conv_bias = weights_dict[bias_path]
    else:
        conv_bias = conv_module.biases

    return load_parameters(
        lambda m: (m.weights, m.biases),
        conv_module,
        (conv_weight, conv_bias),
    )


def load_mamba2(
    module: Mamba2,
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
    permute_conv: bool,
) -> Mamba2:
    in_projection = load_linear(module.in_projection, weights_dict, path / "in_proj")
    out_projection = load_linear(module.out_projection, weights_dict, path / "out_proj")
    conv = _load_conv(module.conv, weights_dict, path, permute_conv)

    skip_connection_weight_path = path / "D"
    if skip_connection_weight_path in weights_dict:
        skip_connection_weight = weights_dict[skip_connection_weight_path]
    else:
        skip_connection_weight = module.skip_connection_weight

    gate_bias_path = path / "z_bias"
    if gate_bias_path in weights_dict:
        gate_bias = weights_dict[gate_bias_path]
    else:
        gate_bias = module.gate_bias

    return load_parameters(
        lambda m: (
            m.in_projection,
            m.out_projection,
            m.conv,
            m.skip_connection_weight,
            m.gate_bias,
        ),
        module,
        (in_projection, out_projection, conv, skip_connection_weight, gate_bias),
    )


def load_short_conv(
    module: ShortConv,
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
    permute_conv: bool,
) -> ShortConv:
    in_projection = load_linear(module.in_projection, weights_dict, path / "in_proj")
    out_projection = load_linear(module.out_projection, weights_dict, path / "out_proj")
    conv = _load_conv(module.conv, weights_dict, path, permute_conv)

    return load_parameters(
        lambda m: (m.in_projection, m.out_projection, m.conv),
        module,
        (in_projection, out_projection, conv),
    )


def load_transformer_layer(
    module: TransformerLayer,
    weights_dict: Mapping[str, Array],
    mixer_path: ParameterPath,
    mlp_path: ParameterPath,
    mixer_key: str,
    mlp_key: str,
    pre_mixer_norm_key: str,
    pre_mlp_norm_key: str,
    up_proj_key: str,
    gate_proj_key: str,
    down_proj_key: str,
    permute_conv: bool,
) -> TransformerLayer:
    if module.pre_mixer_norm is not None:
        pre_attention_norm = load_rmsnorm(
            module.pre_mixer_norm,
            weights_dict,
            mixer_path / pre_mixer_norm_key,
        )

    else:
        pre_attention_norm = None
    # Load mixer (attention or mamba)
    if isinstance(module.mixer, Attention):
        mixer = load_attention(module.mixer, weights_dict, mixer_path / mixer_key)
    elif isinstance(module.mixer, Mamba2):
        mixer = load_mamba2(module.mixer, weights_dict, mixer_path / mixer_key, permute_conv)
    elif isinstance(module.mixer, ShortConv):
        mixer = load_short_conv(module.mixer, weights_dict, mixer_path / mixer_key, permute_conv)
    else:
        mixer = module.mixer

    if module.post_mixer_norm is not None:
        post_attention_norm = load_rmsnorm(
            module.post_mixer_norm,
            weights_dict,
            mixer_path / "post_attention_layernorm",
        )

        pre_mlp_norm = load_rmsnorm(
            module.pre_mlp_norm,
            weights_dict,
            mlp_path / "pre_feedforward_layernorm",
        )
    else:
        post_attention_norm = None

        pre_mlp_norm = load_rmsnorm(
            module.pre_mlp_norm,
            weights_dict,
            mlp_path / pre_mlp_norm_key,
        )

    mlp = load_mlp(
        module.mlp,
        weights_dict,
        mlp_path / mlp_key,
        up_proj_key,
        gate_proj_key,
        down_proj_key,
    )

    if module.post_mlp_norm is not None:
        post_mlp_norm = load_rmsnorm(
            module.post_mlp_norm,
            weights_dict,
            mlp_path / "post_feedforward_layernorm",
        )
    else:
        post_mlp_norm = None

    return load_parameters(
        lambda m: (
            m.pre_mixer_norm,
            m.mixer,
            m.post_mixer_norm,
            m.pre_mlp_norm,
            m.mlp,
            m.post_mlp_norm,
        ),
        module,
        (
            pre_attention_norm,
            mixer,
            post_attention_norm,
            pre_mlp_norm,
            mlp,
            post_mlp_norm,
        ),
    )


def load_tied_embedding(
    module: TiedEmbedding,
    weights_dict: Mapping[str, Array],
    embedding_path: ParameterPath,
) -> TiedEmbedding:
    weights = weights_dict[embedding_path / "weight"]
    return load_parameters(lambda m: (m.weights,), module, (weights,))


def load_mlx_quantized_tied_embedding(
    module: MLXQuantizedTiedEmbedding,
    weights_dict: Mapping[str, Array],
    embedding_path: ParameterPath,
) -> MLXQuantizedTiedEmbedding:
    qweights = weights_dict[embedding_path / "weight"]
    qscales = weights_dict[embedding_path / "scales"]
    qbiases = weights_dict[embedding_path / "biases"]

    weights = _process_quantized_tensor(
        qweights,
        module.config.embedding_quantization_mode,
        module.activation_precision,
        None,
    )
    scales = qscales.astype(module.activation_precision)
    biases = qbiases.astype(module.activation_precision)

    return load_parameters(lambda m: (m.weights, m.scales, m.biases), module, (weights, scales, biases))


def load_mlx_quantized_untied_embedding(
    module: MLXQuantizedUntiedEmbedding,
    weights_dict: Mapping[str, Array],
    embedding_path: ParameterPath,
    lm_head_path: ParameterPath,
) -> MLXQuantizedUntiedEmbedding:
    input_qweights = weights_dict[embedding_path / "weight"]
    input_qscales = weights_dict[embedding_path / "scales"]
    input_qbiases = weights_dict[embedding_path / "biases"]
    output_qweights = weights_dict[lm_head_path / "weight"]
    output_qscales = weights_dict[lm_head_path / "scales"]
    output_qbiases = weights_dict[lm_head_path / "biases"]

    input_weights = _process_quantized_tensor(
        input_qweights,
        module.config.embedding_quantization_mode,
        module.activation_precision,
        None,
    )
    input_scales = input_qscales.astype(module.activation_precision)
    input_biases = input_qbiases.astype(module.activation_precision)

    output_weights = _process_quantized_tensor(
        output_qweights,
        module.config.embedding_quantization_mode,
        module.activation_precision,
        None,
    )
    output_scales = output_qscales.astype(module.activation_precision)
    output_biases = output_qbiases.astype(module.activation_precision)

    return load_parameters(
        lambda m: (
            m.input_weights,
            m.input_scales,
            m.input_biases,
            m.output_weights,
            m.output_scales,
            m.output_biases,
        ),
        module,
        (input_weights, input_scales, input_biases, output_weights, output_scales, output_biases),
    )


def load_mlx_semi_quantized_untied_embedding(
    module: MLXSemiQuantizedUntiedEmbedding,
    weights_dict: Mapping[str, Array],
    embedding_path: ParameterPath,
    lm_head_path: ParameterPath,
) -> MLXSemiQuantizedUntiedEmbedding:
    input_weights = weights_dict[embedding_path / "weight"]

    output_qweights = weights_dict[lm_head_path / "weight"]
    output_qscales = weights_dict[lm_head_path / "scales"]
    output_qbiases = weights_dict[lm_head_path / "biases"]

    output_weights = _process_quantized_tensor(
        output_qweights,
        module.config.embedding_quantization_mode,
        module.activation_precision,
        None,
    )
    output_scales = output_qscales.astype(module.activation_precision)
    output_biases = output_qbiases.astype(module.activation_precision)

    return load_parameters(
        lambda m: (m.input_weights, m.output_weights, m.output_scales, m.output_biases),
        module,
        (input_weights, output_weights, output_scales, output_biases),
    )


def load_untied_embedding(
    module: UntiedEmbedding,
    weights_dict: Mapping[str, Array],
    embedding_path: ParameterPath,
    lm_head_path: ParameterPath,
) -> UntiedEmbedding:
    input_weights = weights_dict[embedding_path / "weight"]
    output_weights = weights_dict[lm_head_path / "weight"]
    return load_parameters(
        lambda m: (m.input_weights, m.output_weights),
        module,
        (input_weights, output_weights),
    )


def load_huggingface_decoder(
    module: Decoder,
    weights_dict: Mapping[str, Array],
) -> Decoder:
    if any(key.startswith("language_model.") for key in weights_dict):
        base_path = ParameterPath("language_model")
    else:
        base_path = ParameterPath()

    is_llamba_full_precision = any(key.startswith("backbone.") for key in weights_dict)
    is_llamba_mlx = any(key.startswith("embedding.encoder.") for key in weights_dict)
    is_lfm2 = any(key.startswith("model.layers.0.operator_norm.weight") for key in weights_dict)
    if is_llamba_full_precision:
        decoder_path = base_path / "backbone"
        embedding_path = decoder_path / "embedding"
        pre_mixer_norm_key = "input_layernorm"
        mixer_key = {Mamba2Config: "mixer"}
        permute_conv = False
        pre_mlp_norm_key = "post_attention_layernorm"
        mlp_key = "mlp"
        up_proj_key = "up_proj"
        gate_proj_key = "gate_proj"
        down_proj_key = "down_proj"
        alternating_layers = False
        norm_key = "final_layernorm"
        lm_head_path = base_path / "lm_head"
    elif is_llamba_mlx:
        decoder_path = base_path / "model"
        embedding_path = base_path / "embedding.encoder"
        pre_mixer_norm_key = "norm"
        mixer_key = {Mamba2Config: "layer"}
        permute_conv = False
        pre_mlp_norm_key = "norm"
        mlp_key = "layer"
        up_proj_key = "gate_proj"
        gate_proj_key = "in_proj"
        down_proj_key = "out_proj"
        alternating_layers = True
        norm_key = "norm"
        lm_head_path = base_path / "head.linear"
    elif is_lfm2:
        decoder_path = base_path / "model"
        embedding_path = decoder_path / "embed_tokens"
        pre_mixer_norm_key = "operator_norm"
        mixer_key = {ShortConvConfig: "conv", AttentionConfig: "self_attn"}
        permute_conv = isinstance(module.config.embedding_config, MLXQuantizedTiedEmbeddingConfig)
        pre_mlp_norm_key = "ffn_norm"
        mlp_key = "feed_forward"
        up_proj_key = "w3"
        gate_proj_key = "w1"
        down_proj_key = "w2"
        alternating_layers = False
        norm_key = "embedding_norm"
        lm_head_path = base_path / "lm_head"
    else:
        decoder_path = base_path / "model"
        embedding_path = decoder_path / "embed_tokens"
        pre_mixer_norm_key = "input_layernorm"
        mixer_key = {AttentionConfig: "self_attn"}
        permute_conv = False
        pre_mlp_norm_key = "post_attention_layernorm"
        mlp_key = "mlp"
        up_proj_key = "up_proj"
        gate_proj_key = "gate_proj"
        down_proj_key = "down_proj"
        alternating_layers = False
        norm_key = "norm"
        lm_head_path = base_path / "lm_head"

    if isinstance(module.embedding, TiedEmbedding):
        embedding = load_tied_embedding(module.embedding, weights_dict, embedding_path)
    elif isinstance(module.embedding, MLXQuantizedTiedEmbedding):
        embedding = load_mlx_quantized_tied_embedding(module.embedding, weights_dict, embedding_path)
    elif isinstance(module.embedding, MLXQuantizedUntiedEmbedding):
        embedding = load_mlx_quantized_untied_embedding(module.embedding, weights_dict, embedding_path, lm_head_path)
    elif isinstance(module.embedding, MLXSemiQuantizedUntiedEmbedding):
        embedding = load_mlx_semi_quantized_untied_embedding(
            module.embedding,
            weights_dict,
            embedding_path,
            lm_head_path,
        )
    elif isinstance(module.embedding, UntiedEmbedding):
        embedding = load_untied_embedding(module.embedding, weights_dict, embedding_path, lm_head_path)
    else:
        raise TypeError(f"Unsupported embedding type: {type(module.embedding)}")

    decoder_layers = tuple(
        load_transformer_layer(
            layer,
            weights_dict,
            decoder_path / "layers" / ((i * 2) if alternating_layers else i),
            decoder_path / "layers" / ((i * 2 + 1) if alternating_layers else i),
            mixer_key[type(layer.config.mixer_config)],  # type: ignore
            mlp_key,
            pre_mixer_norm_key,
            pre_mlp_norm_key,
            up_proj_key,
            gate_proj_key,
            down_proj_key,
            permute_conv,
        )
        for i, layer in enumerate(module.transformer.layers)
    )
    output_norm = load_rmsnorm(module.transformer.output_norm, weights_dict, decoder_path / norm_key)
    return load_parameters(
        lambda m: (m.embedding, m.transformer.layers, m.transformer.output_norm),
        module,
        (embedding, decoder_layers, output_norm),
    )


def load_huggingface_classifier(
    module: Classifier,
    weights_dict: Mapping[str, Array],
) -> Classifier:
    def load_tied_embedding_local(
        module: TiedEmbedding,
        weights_dict: Mapping[str, Array],
        decoder_path: ParameterPath,
    ) -> TiedEmbedding:
        input_weights = weights_dict[decoder_path / "embeddings" / "tok_embeddings" / "weight"]
        return load_parameters(lambda m: (m.weights,), module, (input_weights,))

    def load_linear_with_reshufling(
        module: LinearBase,
        weights_dict: Mapping[str, Array],
        path: ParameterPath,
    ) -> LinearBase:
        """Loads a linear layer and reshufle some weights in resulting matrix to meet
        requirements of downstream 'split' in MLP layer in attention."""

        assert not module.has_biases, "Expecting no biases in FullPrecisionLinear"
        assert isinstance(module, FullPrecisionLinear), "Expecting FullPrecisionLinear module as input"

        weights = weights_dict[path / "weight"]
        rows, _ = weights.shape
        shuffled_weights = jnp.vstack((weights[rows // 2 :, :], weights[: rows // 2, :]))
        return load_parameters(lambda m: (m.weights, m.biases), module, (shuffled_weights, None))

    def load_attention_local(
        module: Attention,
        weights_dict: Mapping[str, Array],
        path: ParameterPath,
    ) -> Attention:
        qkv_projection = load_linear(
            module.qkv_projection,
            weights_dict,
            path / "Wqkv",
            sublayers_to_fuse=None,
        )
        out_projection = load_linear(module.out_projection, weights_dict, path / "Wo")

        if module.query_norm is not None:
            query_norm = load_rmsnorm(module.query_norm, weights_dict, path / "q_norm")
        else:
            query_norm = None

        if module.key_norm is not None:
            key_norm = load_rmsnorm(module.key_norm, weights_dict, path / "k_norm")
        else:
            key_norm = None

        return load_parameters(
            lambda m: (m.qkv_projection, m.out_projection, m.query_norm, m.key_norm),
            module,
            (qkv_projection, out_projection, query_norm, key_norm),
        )

    def load_mlp_local(module: MLPBase, weights_dict: Mapping[str, Array], path: ParameterPath) -> MLPBase:
        assert isinstance(module, DenseMLP)
        up_projection = load_linear_with_reshufling(
            module.up_projection,
            weights_dict,
            path / "Wi",
        )
        down_projection = load_linear(module.down_projection, weights_dict, path / "Wo")
        return load_parameters(
            lambda m: (m.up_projection, m.down_projection),
            module,
            (up_projection, down_projection),
        )

    def load_transformer_layer_local(
        module: TransformerLayer,
        weights_dict: Mapping[str, Array],
        path: ParameterPath,
    ) -> TransformerLayer:
        if module.pre_mixer_norm is not None:
            pre_attention_norm = load_rmsnorm(
                module.pre_mixer_norm,
                weights_dict,
                path / "attn_norm",
            )
        else:
            pre_attention_norm = None

        assert isinstance(module.mixer, Attention)
        attention = load_attention_local(module.mixer, weights_dict, path / "attn")
        if module.post_mixer_norm is not None:
            post_attention_norm = load_rmsnorm(
                module.post_mixer_norm,
                weights_dict,
                path / "post_attention_layernorm",
            )

            pre_mlp_norm = load_rmsnorm(
                module.pre_mlp_norm,
                weights_dict,
                path / "pre_feedforward_layernorm",
            )
        else:
            post_attention_norm = None

            pre_mlp_norm = load_rmsnorm(
                module.pre_mlp_norm,
                weights_dict,
                path / "mlp_norm",
            )

        mlp = load_mlp_local(module.mlp, weights_dict, path / "mlp")
        if module.post_mlp_norm is not None:
            post_mlp_norm = load_rmsnorm(
                module.post_mlp_norm,
                weights_dict,
                path / "post_feedforward_layernorm",
            )
        else:
            post_mlp_norm = None
        return load_parameters(
            lambda m: (
                m.pre_mixer_norm,
                m.mixer,
                m.post_mixer_norm,
                m.pre_mlp_norm,
                m.mlp,
                m.post_mlp_norm,
            ),
            module,
            (
                pre_attention_norm,
                attention,
                post_attention_norm,
                pre_mlp_norm,
                mlp,
                post_mlp_norm,
            ),
        )

    base_path = ParameterPath()
    decoder_path = base_path / "model"
    head_path = base_path / "head"
    classifier_path = base_path / "classifier"
    assert isinstance(module.embedding, TiedEmbedding)
    embedding = load_tied_embedding_local(module.embedding, weights_dict, decoder_path)
    embedding_norm = load_rmsnorm(module.embedding_norm, weights_dict, base_path / "model" / "embeddings" / "norm")

    decoder_layers = tuple(
        load_transformer_layer_local(layer, weights_dict, decoder_path / "layers" / i)
        for i, layer in enumerate(module.transformer.layers)
    )
    output_norm = load_rmsnorm(module.transformer.output_norm, weights_dict, decoder_path / "final_norm")
    head_dense = load_linear(module.prediction_head.dense, weights_dict, head_path / "dense")
    head_norm = load_rmsnorm(module.prediction_head.norm, weights_dict, head_path / "norm")
    head_readout = load_linear(module.prediction_head.readout, weights_dict, classifier_path)
    return load_parameters(
        lambda m: (
            m.embedding,
            m.embedding_norm,
            m.transformer.layers,
            m.transformer.output_norm,
            m.prediction_head.dense,
            m.prediction_head.norm,
            m.prediction_head.readout,
        ),
        module,
        (
            embedding,
            embedding_norm,
            decoder_layers,
            output_norm,
            head_dense,
            head_norm,
            head_readout,
        ),
    )



import math
from collections.abc import Mapping
from dataclasses import dataclass, replace
from typing import NamedTuple, Self

import equinox as eqx
import jax
import jax.numpy as jnp
from einops import einsum, rearrange
from jax import vmap
from jaxtyping import Array, DTypeLike, Float, Int, PRNGKeyArray

from lalamo.common import ParameterTree, dummy_array
from lalamo.modules.activations import Activation
from lalamo.modules.common import LalamoModule, PositionalEmbeddingSelector
from lalamo.modules.linear import LinearBase, LinearConfig
from lalamo.modules.rope import PositionalEmbeddings

from .common import TokenMixerBase, TokenMixerConfigBase, TokenMixerResult
from .state import Mamba2StateLayer

__all__ = [
    "Mamba2",
    "Mamba2Config",
    "Mamba2Result",
    "SeparableCausalConv",
    "SeparableCausalConvConfig",
]


Mamba2Result = TokenMixerResult[Mamba2StateLayer]


class CausalConvResult(NamedTuple):
    outputs: Float[Array, "*batch suffix_tokens channels"]
    state: Float[Array, "*batch tokens channels"] | None = None


@dataclass(frozen=True)
class SeparableCausalConvConfig:
    precision: DTypeLike
    has_biases: bool

    def random_init(
        self,
        input_dim: int,
        kernel_size: int,
        *,
        key: PRNGKeyArray,
    ) -> "SeparableCausalConv":
        scale = 1 / math.sqrt(kernel_size * input_dim)
        weights = jax.random.uniform(
            key,
            (input_dim, kernel_size),
            minval=-scale,
            maxval=scale,
            dtype=self.precision,
        )
        if self.has_biases:
            biases = jnp.zeros((input_dim,), dtype=self.precision)
        else:
            biases = None
        return SeparableCausalConv(self, weights=weights, biases=biases)

    def empty(
        self,
        input_dim: int,
        kernel_size: int,
    ) -> "SeparableCausalConv":
        weights = dummy_array(
            (input_dim, kernel_size),
            dtype=self.precision,
        )
        if self.has_biases:
            biases = dummy_array((input_dim,), dtype=self.precision)
        else:
            biases = None
        return SeparableCausalConv(self, weights=weights, biases=biases)


class SeparableCausalConv(LalamoModule[SeparableCausalConvConfig]):
    weights: Float[Array, "channels kernel"]
    biases: Float[Array, " channels"] | None

    def __post_init__(self) -> None:
        input_dim, _ = self.weights.shape
        if self.biases is not None:
            (output_dim,) = self.biases.shape
            if output_dim != input_dim:
                raise ValueError(
                    f"Output dimension of biases ({output_dim}) must match input dimension ({input_dim})",
                )

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.precision

    @property
    def input_dim(self) -> int:
        input_dim, _ = self.weights.shape
        return input_dim

    @property
    def kernel_size(self) -> int:
        _, kernel_size = self.weights.shape
        return kernel_size

    @property
    def has_biases(self) -> bool:
        return self.biases is not None

    def __call__(
        self,
        inputs: Float[Array, "suffix_tokens channels"],
        state: Float[Array, "prefix_tokens channels"] | None = None,
        return_updated_state: bool = False,
    ) -> CausalConvResult:
        num_suffix_tokens, input_dim = inputs.shape

        if state is None:
            state = jnp.zeros((self.kernel_size - 1, input_dim), dtype=inputs.dtype)

        required_context = num_suffix_tokens + self.kernel_size - 1

        inputs_with_history = jnp.concatenate([state, inputs], axis=0)
        conv_outputs = jax.lax.conv_general_dilated(
            inputs_with_history[None, -required_context:, :],
            self.weights[:, :, None],
            window_strides=(1,),
            feature_group_count=input_dim,
            padding="VALID",
            dimension_numbers=("NTC", "OTI", "NTC"),
        )

        results = conv_outputs.squeeze(0)
        if self.biases is not None:
            results = results + self.biases

        return CausalConvResult(
            results,
            (inputs_with_history if return_updated_state else None),
        )

    def export_weights(self) -> ParameterTree:
        result: dict[str, Array] = {"weights": self.weights}
        if self.biases is not None:
            result["biases"] = self.biases
        return result

    def import_weights(self, weights: ParameterTree[Array]) -> "SeparableCausalConv":
        assert isinstance(weights, Mapping)
        assert isinstance(weights["weights"], Array)
        if self.biases is not None:
            assert isinstance(weights["biases"], Array)
            biases = weights["biases"]
        else:
            biases = None
        return replace(
            self,
            weights=weights["weights"],
            biases=biases,
        )


@dataclass(frozen=True)
class Mamba2Config(TokenMixerConfigBase):
    in_projection_config: LinearConfig
    out_projection_config: LinearConfig
    conv_config: SeparableCausalConvConfig
    activation: Activation

    kernel_size: int
    num_heads: int
    num_groups: int
    head_dim: int
    state_dim: int
    expansion_factor: int

    has_in_biases: bool
    has_out_biases: bool

    @property
    def inner_dim(self) -> int:
        return self.num_heads * self.head_dim

    @property
    def rope_dim(self) -> None:
        return None

    def random_init(
        self,
        model_dim: int,
        *,
        key: PRNGKeyArray,
    ) -> "Mamba2":
        in_key, out_key, conv_key, skip_key = jax.random.split(key, 4)

        in_projection = self.in_projection_config.random_init(
            input_dim=model_dim,
            output_dims=(
                self.inner_dim + 2 * self.num_groups * self.state_dim,
                self.inner_dim,
                self.num_heads,
            ),
            has_biases=self.has_in_biases,
            key=in_key,
        )

        out_projection = self.out_projection_config.random_init(
            self.inner_dim,
            (model_dim,),
            has_biases=self.has_out_biases,
            key=out_key,
        )

        conv_channels = self.inner_dim + 2 * self.num_groups * self.state_dim
        conv = self.conv_config.random_init(conv_channels, self.kernel_size, key=conv_key)

        skip_connection_weight = jax.random.normal(
            skip_key,
            (self.num_heads,),
            dtype=in_projection.activation_precision,
        )

        gate_bias = jnp.zeros((self.inner_dim,), dtype=in_projection.activation_precision)

        return Mamba2(
            self,
            in_projection=in_projection,
            conv=conv,
            out_projection=out_projection,
            skip_connection_weight=skip_connection_weight,
            gate_bias=gate_bias,
            num_heads=self.num_heads,
            num_groups=self.num_groups,
            head_dim=self.head_dim,
            state_dim=self.state_dim,
        )

    def empty(
        self,
        model_dim: int,
    ) -> "Mamba2":
        in_projection = self.in_projection_config.empty(
            input_dim=model_dim,
            output_dims=(
                self.inner_dim + 2 * self.num_groups * self.state_dim,
                self.inner_dim,
                self.num_heads,
            ),
            has_biases=self.has_in_biases,
        )

        out_projection = self.out_projection_config.empty(
            self.inner_dim,
            (model_dim,),
            has_biases=self.has_out_biases,
        )

        conv_channels = self.inner_dim + 2 * self.num_groups * self.state_dim
        conv = self.conv_config.empty(conv_channels, self.kernel_size)

        skip_connection_weight = dummy_array((self.num_heads,), in_projection.activation_precision)
        gate_bias = dummy_array((self.inner_dim,), in_projection.activation_precision)

        return Mamba2(
            self,
            in_projection=in_projection,
            conv=conv,
            out_projection=out_projection,
            skip_connection_weight=skip_connection_weight,
            gate_bias=gate_bias,
            num_heads=self.num_heads,
            num_groups=self.num_groups,
            head_dim=self.head_dim,
            state_dim=self.state_dim,
        )


class Mamba2(TokenMixerBase[Mamba2Config, Mamba2StateLayer]):
    in_projection: LinearBase
    conv: SeparableCausalConv
    out_projection: LinearBase

    skip_connection_weight: Float[Array, " heads"]
    gate_bias: Float[Array, " inner_channels"]

    num_heads: int = eqx.field(static=True)
    num_groups: int = eqx.field(static=True)
    head_dim: int = eqx.field(static=True)
    state_dim: int = eqx.field(static=True)

    @property
    def activation_precision(self) -> DTypeLike:
        return self.in_projection.activation_precision

    @property
    def model_dim(self) -> int:
        return self.in_projection.input_dim

    @property
    def inner_dim(self) -> int:
        return self.num_heads * self.head_dim

    @property
    def positional_embedding_selector(self) -> PositionalEmbeddingSelector:
        return PositionalEmbeddingSelector.NONE

    def __post_init__(self) -> None:
        if self.skip_connection_weight.shape != (self.num_heads,):
            raise ValueError(
                f"Skip connection weight must have shape (num_heads,) = ({self.num_heads},), "
                f"got {self.skip_connection_weight.shape}",
            )
        if self.gate_bias.shape != (self.inner_dim,):
            raise ValueError(
                f"Gate bias must have shape (inner_dim,) = ({self.inner_dim},), got {self.gate_bias.shape}",
            )
        if self.num_heads % self.num_groups != 0:
            raise ValueError(
                f"Number of value heads ({self.num_heads}) must be divisible by number of groups ({self.num_groups})",
            )

    def _scan(
        self,
        hidden_states: Float[Array, "suffix_tokens heads head_channels"],
        input_projection: Float[Array, "suffix_tokens groups state_channels"],
        output_projection: Float[Array, "suffix_tokens groups state_channels"],
        time_delta_log: Float[Array, "suffix_tokens heads"],
        initial_state: Float[Array, "heads head_channels state_channels"],
        num_steps: Int[Array, ""] | int,
    ) -> tuple[
        Float[Array, "suffix_tokens heads head_channels"],
        Float[Array, "heads head_channels state_channels"],
    ]:
        def scan_fn(
            index_and_carry_state: tuple[Int[Array, ""], Float[Array, "heads head_channels state_channels"]],
            step_inputs: tuple[
                Float[Array, "heads head_channels"],
                Float[Array, "groups state_channels"],
                Float[Array, "groups state_channels"],
                Float[Array, " heads"],
            ],
        ) -> tuple[
            tuple[Int[Array, ""], Float[Array, "heads head_channels state_channels"]],
            Float[Array, "heads head_channels"],
        ]:
            index, carry_state = index_and_carry_state
            hidden_state_t, input_proj_t, output_proj_t, time_delta_log_t = step_inputs
            dt = jax.nn.softplus(time_delta_log_t)[:, None]
            heads_per_group = self.num_heads // self.num_groups

            hidden_grouped = rearrange(
                hidden_state_t,
                "(groups heads) head_channels -> groups heads head_channels",
                groups=self.num_groups,
                heads=heads_per_group,
            )
            x_norm_grouped = hidden_grouped / (
                dt.reshape(self.num_heads)[
                    rearrange(
                        jnp.arange(self.num_heads),
                        "(groups heads)-> groups heads",
                        groups=self.num_groups,
                        heads=heads_per_group,
                    )
                ][:, :, None]
                + 1e-8
            )

            decay = jnp.exp(-dt)[:, :, None]
            mix = dt[:, :, None]
            decay_group = rearrange(
                decay,
                "(groups heads) 1 1 -> groups heads 1 1",
                groups=self.num_groups,
                heads=heads_per_group,
            )
            mix_group = rearrange(
                mix,
                "(groups heads) 1 1 -> groups heads 1 1",
                groups=self.num_groups,
                heads=heads_per_group,
            )

            input_contribution_group = mix_group * x_norm_grouped[:, :, :, None] * input_proj_t[:, None, None, :]
            carry_state_group = rearrange(
                carry_state,
                "(groups heads) head_channels state_channels -> groups heads head_channels state_channels",
                groups=self.num_groups,
                heads=heads_per_group,
            )
            updated_state_group = decay_group * carry_state_group + input_contribution_group

            output_group = einsum(
                updated_state_group,
                output_proj_t,
                "groups heads head_channels state_channels, groups state_channels -> groups heads head_channels",
            )
            updated_state = rearrange(
                updated_state_group,
                "groups heads head_channels state_channels -> (groups heads) head_channels state_channels",
            )
            output_t = rearrange(output_group, "groups heads head_channels -> (groups heads) head_channels")

            propagated_state = jax.lax.cond(index < num_steps, lambda: updated_state, lambda: carry_state)

            return (index + 1, propagated_state), output_t

        (_, final_state), outputs = jax.lax.scan(
            scan_fn,
            (jnp.zeros((), dtype=jnp.int32), initial_state),
            (hidden_states, input_projection, output_projection, time_delta_log),
        )

        return outputs, final_state

    @eqx.filter_jit
    def __call__(
        self,
        inputs: Float[Array, "suffix_tokens channels"],
        positional_embeddings: PositionalEmbeddings | None,
        state: Mamba2StateLayer | None = None,
        return_updated_state: bool = False,
        length_without_padding: Int[Array, ""] | int | None = None,
    ) -> Mamba2Result:
        if positional_embeddings is not None:
            raise ValueError("Positional embeddings are not supported for Mamba2.")

        conv_inputs, gate_values, time_delta_log = vmap(self.in_projection)(inputs)

        if state is None:
            state = Mamba2StateLayer.init(
                self.config.kernel_size,
                self.inner_dim,
                self.num_heads,
                self.num_groups,
                self.head_dim,
                self.state_dim,
                self.activation_precision,
            )

        conv_output, updated_conv_state = self.conv(
            conv_inputs,
            state.conv_state,
            return_updated_state=return_updated_state,
        )
        conv_activated = self.config.activation(conv_output)

        x_channels, input_proj_channels, output_proj_channels = jnp.split(
            conv_activated,
            [
                self.inner_dim,
                self.inner_dim + self.num_groups * self.state_dim,
            ],
            axis=-1,
        )

        hidden_states = rearrange(
            x_channels,
            "suffix_tokens (heads head_channels) -> suffix_tokens heads head_channels",
            heads=self.num_heads,
        )
        input_projection = rearrange(
            input_proj_channels,
            "suffix_tokens (groups state_channels) -> suffix_tokens groups state_channels",
            groups=self.num_groups,
        )
        output_projection = rearrange(
            output_proj_channels,
            "suffix_tokens (groups state_channels) -> suffix_tokens groups state_channels",
            groups=self.num_groups,
        )
        time_delta_log = rearrange(
            time_delta_log,
            "suffix_tokens heads -> suffix_tokens heads",
            heads=self.num_heads,
        )

        if length_without_padding is None:
            length_without_padding, _ = inputs.shape

        ssm_outputs, final_ssm_state = self._scan(
            hidden_states,
            input_projection,
            output_projection,
            time_delta_log,
            state.ssm_state,
            length_without_padding,
        )

        skip_contribution = self.skip_connection_weight[None, :, None] * hidden_states
        ssm_outputs = ssm_outputs + skip_contribution

        ssm_outputs = rearrange(
            ssm_outputs,
            "suffix_tokens heads head_channels -> suffix_tokens (heads head_channels)",
        )

        gated_outputs = ssm_outputs * jax.nn.silu(gate_values + self.gate_bias)

        (outputs,) = vmap(self.out_projection)(gated_outputs)

        if return_updated_state:
            assert updated_conv_state is not None
            updated_state = Mamba2StateLayer(updated_conv_state, final_ssm_state)
        else:
            updated_state = None

        return Mamba2Result(
            outputs=outputs,
            state=updated_state,
        )

    def init_static_state(self, capacity: int) -> Mamba2StateLayer:  # noqa: ARG002
        return Mamba2StateLayer.init(
            self.config.kernel_size,
            self.inner_dim,
            self.num_heads,
            self.num_groups,
            self.head_dim,
            self.state_dim,
            self.activation_precision,
        )

    def export_weights(self) -> ParameterTree:
        return {
            "in_projection": self.in_projection.export_weights(),
            "out_projection": self.out_projection.export_weights(),
            "conv": self.conv.export_weights(),
            "skip_connection_weight": self.skip_connection_weight,
            "gate_bias": self.gate_bias,
        }

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self:
        assert isinstance(weights, Mapping)
        assert isinstance(weights["in_projection"], Mapping)
        assert isinstance(weights["out_projection"], Mapping)
        assert isinstance(weights["conv"], Mapping)
        assert isinstance(weights["skip_connection_weight"], Array)
        assert isinstance(weights["gate_bias"], Array)

        return replace(
            self,
            in_projection=self.in_projection.import_weights(weights["in_projection"]),
            out_projection=self.out_projection.import_weights(weights["out_projection"]),
            conv=self.conv.import_weights(weights["conv"]),
            skip_connection_weight=weights["skip_connection_weight"],
            gate_bias=weights["gate_bias"],
        )



from lalamo.model_import.decoder_configs import HFQwen2Config, HFQwen3Config
from lalamo.quantization import QuantizationMode

from .common import ConfigMap, FileSpec, ModelSpec, UseCase, WeightsType

__all__ = ["QWEN_MODELS"]


QWEN25 = [
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5",
        name="Qwen2.5-0.5B-Instruct",
        size="0.5B",
        quantization=None,
        repo="Qwen/Qwen2.5-0.5B-Instruct",
        config_type=HFQwen2Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5",
        name="Qwen2.5-1.5B-Instruct",
        size="1.5B",
        quantization=None,
        repo="Qwen/Qwen2.5-1.5B-Instruct",
        config_type=HFQwen2Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5",
        name="Qwen2.5-3B-Instruct",
        size="3B",
        quantization=None,
        repo="Qwen/Qwen2.5-3B-Instruct",
        config_type=HFQwen2Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5",
        name="Qwen2.5-7B-Instruct",
        size="7B",
        quantization=None,
        repo="Qwen/Qwen2.5-7B-Instruct",
        config_type=HFQwen2Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5",
        name="Qwen2.5-14B-Instruct",
        size="14B",
        quantization=None,
        repo="Qwen/Qwen2.5-14B-Instruct",
        config_type=HFQwen2Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5",
        name="Qwen2.5-32B-Instruct",
        size="32B",
        quantization=None,
        repo="Qwen/Qwen2.5-32B-Instruct",
        config_type=HFQwen2Config,
    ),
]


QWEN25_CODER = [
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-0.5B-Instruct",
        size="0.5B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-0.5B-Instruct",
        config_type=HFQwen2Config,
        use_cases=(UseCase.CODE,),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-1.5B-Instruct",
        size="1.5B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-1.5B-Instruct",
        config_type=HFQwen2Config,
        use_cases=(UseCase.CODE,),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-3B-Instruct",
        size="3B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-3B-Instruct",
        config_type=HFQwen2Config,
        use_cases=(UseCase.CODE,),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-7B-Instruct",
        size="7B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-7B-Instruct",
        config_type=HFQwen2Config,
        use_cases=(UseCase.CODE,),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-14B-Instruct",
        size="14B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-14B-Instruct",
        config_type=HFQwen2Config,
        use_cases=(UseCase.CODE,),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-32B-Instruct",
        size="32B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-32B-Instruct",
        config_type=HFQwen2Config,
        use_cases=(UseCase.CODE,),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-32B-Instruct",
        size="32B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-32B-Instruct",
        config_type=HFQwen2Config,
        use_cases=(UseCase.CODE,),
    ),
]


QWEN3 = [
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-0.6B",
        size="0.6B",
        quantization=None,
        repo="Qwen/Qwen3-0.6B",
        config_type=HFQwen3Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-0.6B-MLX-4bit",
        size="0.6B",
        quantization=QuantizationMode.UINT4,
        repo="Qwen/Qwen3-0.6B-MLX-4bit",
        config_type=HFQwen3Config,
        configs=ConfigMap(
            tokenizer=FileSpec("tokenizer.json", "Qwen/Qwen3-0.6B"),
            tokenizer_config=FileSpec("tokenizer_config.json", "Qwen/Qwen3-0.6B"),
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3-0.6B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-1.7B",
        size="1.7B",
        quantization=None,
        repo="Qwen/Qwen3-1.7B",
        config_type=HFQwen3Config,
        weights_type=WeightsType.SAFETENSORS,
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-4B",
        size="4B",
        quantization=None,
        repo="Qwen/Qwen3-4B",
        config_type=HFQwen3Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-4B-AWQ",
        size="4B",
        quantization=QuantizationMode.UINT4,
        repo="Qwen/Qwen3-4B-AWQ",
        config_type=HFQwen3Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-4B-MLX-4bit",
        size="4B",
        quantization=QuantizationMode.UINT4,
        repo="Qwen/Qwen3-4B-MLX-4bit",
        config_type=HFQwen3Config,
        configs=ConfigMap(
            tokenizer=FileSpec("tokenizer.json", "Qwen/Qwen3-4B"),
            tokenizer_config=FileSpec("tokenizer_config.json", "Qwen/Qwen3-4B"),
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3-4B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-8B",
        size="8B",
        quantization=None,
        repo="Qwen/Qwen3-8B",
        config_type=HFQwen3Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-8B-AWQ",
        size="8B",
        quantization=QuantizationMode.UINT4,
        repo="Qwen/Qwen3-8B-AWQ",
        config_type=HFQwen3Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-8B-MLX-4bit",
        size="8B",
        quantization=QuantizationMode.UINT4,
        repo="Qwen/Qwen3-8B-MLX-4bit",
        config_type=HFQwen3Config,
        configs=ConfigMap(
            tokenizer=FileSpec("tokenizer.json", "Qwen/Qwen3-8B"),
            tokenizer_config=FileSpec("tokenizer_config.json", "Qwen/Qwen3-8B"),
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3-8B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-14B",
        size="14B",
        quantization=None,
        repo="Qwen/Qwen3-14B",
        config_type=HFQwen3Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-14B-AWQ",
        size="14B",
        quantization=None,
        repo="Qwen/Qwen3-14B-AWQ",
        config_type=HFQwen3Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-32B",
        size="32B",
        quantization=None,
        repo="Qwen/Qwen3-32B",
        config_type=HFQwen3Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-32B-AWQ",
        size="32B",
        quantization=QuantizationMode.UINT4,
        repo="Qwen/Qwen3-32B-AWQ",
        config_type=HFQwen3Config,
    ),
]


QWEN_MODELS = QWEN25 + QWEN25_CODER + QWEN3



from lalamo.model_import.decoder_configs import (
    HFGemma2Config,
    HFGemma3Config,
    HFGemma3TextConfig,
)
from lalamo.quantization import QuantizationMode

from .common import ConfigMap, FileSpec, ModelSpec, WeightsType

__all__ = ["GEMMA_MODELS"]

GEMMA2 = [
    ModelSpec(
        vendor="Google",
        family="Gemma-2",
        name="Gemma-2-2B-Instruct",
        size="2B",
        quantization=None,
        repo="google/gemma-2-2b-it",
        config_type=HFGemma2Config,
    ),
]

GEMMA3 = [
    ModelSpec(
        vendor="Google",
        family="Gemma-3",
        name="Gemma-3-1B-Instruct",
        size="1B",
        quantization=None,
        repo="google/gemma-3-1b-it",
        config_type=HFGemma3TextConfig,
        weights_type=WeightsType.SAFETENSORS,
    ),
    ModelSpec(
        vendor="Google",
        family="Gemma-3",
        name="Gemma-3-1B-Instruct-4bit",
        size="1B",
        quantization=QuantizationMode.UINT4,
        repo="mlx-community/gemma-3-1b-it-4bit",
        config_type=HFGemma3TextConfig,
        weights_type=WeightsType.SAFETENSORS,
        configs=ConfigMap(generation_config=FileSpec("generation_config.json", "google/gemma-3-1b-it")),
    ),
    ModelSpec(
        vendor="Google",
        family="Gemma-3",
        name="Gemma-3-1B-Instruct-8bit",
        size="1B",
        quantization=QuantizationMode.UINT8,
        repo="mlx-community/gemma-3-1b-it-8bit",
        config_type=HFGemma3TextConfig,
        weights_type=WeightsType.SAFETENSORS,
        configs=ConfigMap(generation_config=FileSpec("generation_config.json", "google/gemma-3-1b-it")),
    ),
    ModelSpec(
        vendor="Google",
        family="Gemma-3",
        name="Gemma-3-4B-Instruct",
        size="4B",
        quantization=None,
        repo="google/gemma-3-4b-it",
        config_type=HFGemma3Config,
        weights_type=WeightsType.SAFETENSORS,
    ),
    ModelSpec(
        vendor="Google",
        family="Gemma-3",
        name="Gemma-3-4B-Instruct-4bit",
        size="4B",
        quantization=QuantizationMode.UINT4,
        repo="mlx-community/gemma-3-4b-it-4bit",
        config_type=HFGemma3Config,
        weights_type=WeightsType.SAFETENSORS,
        configs=ConfigMap(generation_config=FileSpec("generation_config.json", "google/gemma-3-4b-it")),
    ),
    ModelSpec(
        vendor="Google",
        family="Gemma-3",
        name="Gemma-3-4B-Instruct-8bit",
        size="4B",
        quantization=QuantizationMode.UINT8,
        repo="mlx-community/gemma-3-4b-it-8bit",
        config_type=HFGemma3Config,
        weights_type=WeightsType.SAFETENSORS,
        configs=ConfigMap(generation_config=FileSpec("generation_config.json", "google/gemma-3-4b-it")),
    ),
    ModelSpec(
        vendor="Google",
        family="Gemma-3",
        name="Gemma-3-27B-Instruct",
        size="27B",
        quantization=None,
        repo="google/gemma-3-27b-it",
        config_type=HFGemma3Config,
        weights_type=WeightsType.SAFETENSORS,
    ),
    ModelSpec(
        vendor="Google",
        family="Gemma-3",
        name="Gemma-3-27B-Instruct-4bit",
        size="27B",
        quantization=QuantizationMode.UINT4,
        repo="mlx-community/gemma-3-27b-it-4bit",
        config_type=HFGemma3Config,
        weights_type=WeightsType.SAFETENSORS,
        configs=ConfigMap(generation_config=FileSpec("generation_config.json", "google/gemma-3-27b-it")),
    ),
    ModelSpec(
        vendor="Google",
        family="Gemma-3",
        name="Gemma-3-27B-Instruct-8bit",
        size="27B",
        quantization=QuantizationMode.UINT8,
        repo="mlx-community/gemma-3-27b-it-8bit",
        config_type=HFGemma3Config,
        weights_type=WeightsType.SAFETENSORS,
        configs=ConfigMap(generation_config=FileSpec("generation_config.json", "google/gemma-3-27b-it")),
    ),
]


GEMMA_MODELS = GEMMA2 + GEMMA3



from lalamo.model_import.decoder_configs.huggingface.llama import HFLlamaConfig
from lalamo.model_import.model_specs.common import ModelSpec, WeightsType

EXAMPLE_JSON = {
    "vendor": "Meta",
    "family": "Llama-3.2",
    "name": "Llama-3.2-1B-Instruct",
    "size": "1B",
    "quantization": None,
    "repo": "meta-llama/Llama-3.2-1B-Instruct",
    "config_type": "HFLlamaConfig",
    "output_parser_regex": None,
    "system_role_name": "system",
    "user_role_name": "user",
    "assistant_role_name": "assistant",
    "tool_role_name": "tool",
    "weights_type": "safetensors",
    "use_cases": [],
}


def test_deserialization() -> None:
    model_spec = ModelSpec.from_json(EXAMPLE_JSON)
    assert model_spec.vendor == "Meta"
    assert model_spec.family == "Llama-3.2"
    assert model_spec.name == "Llama-3.2-1B-Instruct"
    assert model_spec.size == "1B"
    assert model_spec.quantization is None
    assert model_spec.repo == "meta-llama/Llama-3.2-1B-Instruct"
    assert model_spec.config_type is HFLlamaConfig
    assert model_spec.output_parser_regex is None
    assert model_spec.system_role_name == "system"
    assert model_spec.user_role_name == "user"
    assert model_spec.assistant_role_name == "assistant"
    assert model_spec.tool_role_name == "tool"
    assert model_spec.weights_type == WeightsType.SAFETENSORS
    assert model_spec.use_cases == ()


def test_consistency() -> None:
    spec = ModelSpec(
        vendor="Meta",
        family="Llama-3.2",
        name="Llama-3.2-1B-Instruct",
        size="1B",
        quantization=None,
        repo="meta-llama/Llama-3.2-1B-Instruct",
        config_type=HFLlamaConfig,
        use_cases=tuple(),
    )

    assert ModelSpec.from_json(spec.to_json()) == spec



from typing import Any

import jax.numpy as jnp
import pytest

from lalamo.common import unflatten_parameters
from lalamo.utils import MapDictValues, MapSequence


def test_unflatten_simple_dict() -> None:
    flat = {"a": jnp.array([1]), "b": jnp.array([2])}
    result = unflatten_parameters(flat)
    assert isinstance(result, MapDictValues)
    assert "a" in result and "b" in result
    assert jnp.array_equal(result["a"], jnp.array([1]))  # type: ignore
    assert jnp.array_equal(result["b"], jnp.array([2]))  # type: ignore


def test_unflatten_nested_dict() -> None:
    flat = {
        "layer.weight": jnp.array([1]),
        "layer.bias": jnp.array([2]),
    }
    result = unflatten_parameters(flat)
    assert isinstance(result, MapDictValues)
    assert isinstance(result["layer"], MapDictValues)  # type: ignore
    assert jnp.array_equal(result["layer"]["weight"], jnp.array([1]))  # type: ignore
    assert jnp.array_equal(result["layer"]["bias"], jnp.array([2]))  # type: ignore


def test_unflatten_to_list() -> None:
    flat = {"0": jnp.array([1]), "1": jnp.array([2])}
    result = unflatten_parameters(flat)
    assert isinstance(result, MapSequence)
    assert len(result) == 2
    assert jnp.array_equal(result[0], jnp.array([1]))  # type: ignore
    assert jnp.array_equal(result[1], jnp.array([2]))  # type: ignore


def test_unflatten_mixed_nested() -> None:
    flat = {
        "layers.0.weight": jnp.array([1]),
        "layers.0.bias": jnp.array([2]),
        "layers.1.weight": jnp.array([3]),
        "layers.1.bias": jnp.array([4]),
    }
    result = unflatten_parameters(flat)
    assert isinstance(result, MapDictValues)
    assert isinstance(result["layers"], MapSequence)  # type: ignore
    assert len(result["layers"]) == 2  # type: ignore
    assert isinstance(result["layers"][0], MapDictValues)  # type: ignore
    assert isinstance(result["layers"][1], MapDictValues)  # type: ignore
    assert jnp.array_equal(result["layers"][0]["weight"], jnp.array([1]))  # type: ignore
    assert jnp.array_equal(result["layers"][0]["bias"], jnp.array([2]))  # type: ignore
    assert jnp.array_equal(result["layers"][1]["weight"], jnp.array([3]))  # type: ignore
    assert jnp.array_equal(result["layers"][1]["bias"], jnp.array([4]))  # type: ignore


def test_unflatten_empty_dict() -> None:
    flat: dict[str, Any] = {}
    result = unflatten_parameters(flat)
    assert isinstance(result, MapDictValues)
    assert len(result) == 0


def test_unflatten_single_level_numeric() -> None:
    flat = {
        "0.weight": jnp.array([1]),
        "1.weight": jnp.array([2]),
    }
    result = unflatten_parameters(flat)
    assert isinstance(result, MapSequence)
    assert len(result) == 2
    assert isinstance(result[0], MapDictValues)
    assert isinstance(result[1], MapDictValues)
    assert jnp.array_equal(result[0]["weight"], jnp.array([1]))  # type: ignore
    assert jnp.array_equal(result[1]["weight"], jnp.array([2]))  # type: ignore


def test_unflatten_deep_nesting() -> None:
    flat = {
        "model.layers.0.attention.weight": jnp.array([1]),
        "model.layers.0.attention.bias": jnp.array([2]),
        "model.layers.1.attention.weight": jnp.array([3]),
        "model.layers.1.attention.bias": jnp.array([4]),
    }
    result = unflatten_parameters(flat)
    assert isinstance(result, MapDictValues)
    assert isinstance(result["model"], MapDictValues)  # type: ignore
    assert isinstance(result["model"]["layers"], MapSequence)  # type: ignore
    assert len(result["model"]["layers"]) == 2  # type: ignore
    assert isinstance(result["model"]["layers"][0]["attention"], MapDictValues)  # type: ignore
    assert isinstance(result["model"]["layers"][1]["attention"], MapDictValues)  # type: ignore
    assert jnp.array_equal(result["model"]["layers"][0]["attention"]["weight"], jnp.array([1]))  # type: ignore
    assert jnp.array_equal(result["model"]["layers"][0]["attention"]["bias"], jnp.array([2]))  # type: ignore
    assert jnp.array_equal(result["model"]["layers"][1]["attention"]["weight"], jnp.array([3]))  # type: ignore
    assert jnp.array_equal(result["model"]["layers"][1]["attention"]["bias"], jnp.array([4]))  # type: ignore


def test_unflatten_invalid_numeric_sequence() -> None:
    flat = {"0": jnp.array([1]), "2": jnp.array([2])}  # Missing "1"
    with pytest.raises(AssertionError):
        unflatten_parameters(flat)



from lalamo.model_import.decoder_configs import HFLlambaConfig
from lalamo.quantization import QuantizationMode

from .common import ConfigMap, FileSpec, ModelSpec

__all__ = ["LLAMBA_MODELS"]

LLAMBA_MODELS = [
    ModelSpec(
        vendor="Cartesia",
        family="Llamba",
        name="Llamba-1B",
        size="1B",
        quantization=None,
        repo="cartesia-ai/Llamba-1B",
        config_type=HFLlambaConfig,
        configs=ConfigMap(
            tokenizer=FileSpec("tokenizer.json", "meta-llama/Llama-3.2-1B-Instruct"),
            tokenizer_config=FileSpec("tokenizer_config.json", "meta-llama/Llama-3.2-1B-Instruct"),
            generation_config=FileSpec("generation_config.json", "meta-llama/Llama-3.2-1B-Instruct"),
        ),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Cartesia",
        family="Llamba",
        name="Llamba-1B-4bit-mlx",
        size="1B",
        quantization=QuantizationMode.UINT4,
        repo="cartesia-ai/Llamba-1B-4bit-mlx",
        config_type=HFLlambaConfig,
        configs=ConfigMap(
            model_config=FileSpec("config.json", "cartesia-ai/Llamba-1B"),
            tokenizer=FileSpec("tokenizer.json", "meta-llama/Llama-3.2-1B-Instruct"),
            tokenizer_config=FileSpec("tokenizer_config.json", "meta-llama/Llama-3.2-1B-Instruct"),
            generation_config=FileSpec("generation_config.json", "meta-llama/Llama-3.2-1B-Instruct"),
        ),
        use_cases=tuple(),
    ),
]



from abc import abstractmethod
from collections.abc import Mapping
from dataclasses import dataclass, replace
from typing import Self

import equinox as eqx
import jax
import jax.numpy as jnp
from einops import rearrange
from jaxtyping import Array, DTypeLike, Float, Int, PRNGKeyArray

from lalamo.common import ParameterTree, dummy_array
from lalamo.quantization import QuantizationMode, dynamically_quantize_activations, quantize_weights
from lalamo.utils import jax_uint4_to_packed_uint8, jax_uint8_to_unpacked_uint4

from .common import (
    LalamoModule,
    register_config_union,
)
from .utils import apply_soft_capping

__all__ = [
    "EmbeddingBase",
    "EmbeddingConfig",
    "MLXQuantizedTiedEmbedding",
    "MLXQuantizedTiedEmbeddingConfig",
    "MLXQuantizedUntiedEmbedding",
    "MLXQuantizedUntiedEmbeddingConfig",
    "MLXSemiQuantizedUntiedEmbedding",
    "MLXSemiQuantizedUntiedEmbeddingConfig",
    "QuantizedTiedEmbedding",
    "QuantizedTiedEmbeddingConfig",
    "TiedEmbedding",
    "TiedEmbeddingConfig",
    "UntiedEmbedding",
    "UntiedEmbeddingConfig",
]


@dataclass(frozen=True)
class EmbeddingConfigBase:
    input_scale: float | None
    logit_soft_cap: float | None

    @abstractmethod
    def random_init(
        self,
        vocab_size: int,
        model_dim: int,
        *,
        key: PRNGKeyArray,
    ) -> "EmbeddingBase": ...

    @abstractmethod
    def empty(
        self,
        vocab_size: int,
        model_dim: int,
    ) -> "EmbeddingBase": ...


class EmbeddingBase[ConfigT: EmbeddingConfigBase](LalamoModule[ConfigT]):
    @abstractmethod
    def _prepare_input_weights(self) -> Float[Array, "vocabulary channels"]: ...

    @abstractmethod
    def _prepare_output_weights(self) -> Float[Array, "vocabulary channels"]: ...

    @property
    @abstractmethod
    def vocab_size(self) -> int: ...

    @property
    @abstractmethod
    def model_dim(self) -> int: ...

    @eqx.filter_jit
    def embed(self, x: Int[Array, " tokens"]) -> Float[Array, "tokens channels"]:
        result = self._prepare_input_weights()[x]
        if self.config.input_scale is not None:
            result = result * jnp.array(self.config.input_scale, dtype=result.dtype)
        return result

    @eqx.filter_jit
    def readout(self, x: Float[Array, " channels"]) -> Float[Array, " vocabulary"]:
        logits = self._prepare_output_weights() @ x
        if self.config.logit_soft_cap is not None:
            logits = apply_soft_capping(logits, self.config.logit_soft_cap)
        return logits


@dataclass(frozen=True)
class TiedEmbeddingConfig(EmbeddingConfigBase):
    precision: DTypeLike

    def random_init(
        self,
        vocab_size: int,
        model_dim: int,
        *,
        key: PRNGKeyArray,
    ) -> "TiedEmbedding":
        weights = jax.random.normal(key, (vocab_size, model_dim), dtype=self.precision)
        return TiedEmbedding(config=self, weights=weights)

    def empty(
        self,
        vocab_size: int,
        model_dim: int,
    ) -> "TiedEmbedding":
        weights = dummy_array((vocab_size, model_dim), dtype=self.precision)
        return TiedEmbedding(config=self, weights=weights)


class TiedEmbedding(EmbeddingBase[TiedEmbeddingConfig]):
    weights: Float[Array, "vocabulary channels"]

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.precision

    def __post_init__(self) -> None:
        if self.config.precision != self.weights.dtype:
            raise ValueError(
                f"Embedding dtype {self.weights.dtype} does not match the specified precision {self.config.precision}",
            )

    @property
    def model_dim(self) -> int:
        _, model_dim = self.weights.shape
        return model_dim

    @property
    def vocab_size(self) -> int:
        vocab_size, _ = self.weights.shape
        return vocab_size

    def _prepare_input_weights(self) -> Float[Array, "vocabulary channels"]:
        return self.weights

    def _prepare_output_weights(self) -> Float[Array, "vocabulary channels"]:
        return self.weights

    def export_weights(self) -> ParameterTree:
        return {"weights": self.weights}

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self:
        assert isinstance(weights, Mapping)
        return replace(self, weights=weights["weights"])


@dataclass(frozen=True)
class UntiedEmbeddingConfig(EmbeddingConfigBase):
    precision: DTypeLike

    def random_init(
        self,
        vocab_size: int,
        model_dim: int,
        *,
        key: PRNGKeyArray,
    ) -> "UntiedEmbedding":
        input_key, output_key = jax.random.split(key)
        input_weights = jax.random.normal(input_key, (vocab_size, model_dim), dtype=self.precision)
        output_weights = jax.random.normal(output_key, (vocab_size, model_dim), dtype=self.precision)
        return UntiedEmbedding(
            config=self,
            input_weights=input_weights,
            output_weights=output_weights,
        )

    def empty(
        self,
        vocab_size: int,
        model_dim: int,
    ) -> "UntiedEmbedding":
        input_weights = dummy_array((vocab_size, model_dim), dtype=self.precision)
        output_weights = dummy_array((vocab_size, model_dim), dtype=self.precision)
        return UntiedEmbedding(
            config=self,
            input_weights=input_weights,
            output_weights=output_weights,
        )


class UntiedEmbedding(EmbeddingBase[UntiedEmbeddingConfig]):
    input_weights: Float[Array, "vocabulary channels"]
    output_weights: Float[Array, "channels vocabulary"]

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.precision

    @property
    def model_dim(self) -> int:
        _, model_dim = self.input_weights.shape
        return model_dim

    @property
    def vocab_size(self) -> int:
        vocab_size, _ = self.input_weights.shape
        return vocab_size

    def __post_init__(self) -> None:
        if self.config.precision != self.input_weights.dtype:
            raise ValueError(
                f"Embedding dtype {self.input_weights.dtype} does not match",
                f" the specified precision {self.config.precision}",
            )
        if self.config.precision != self.output_weights.dtype:
            raise ValueError(
                f"Embedding dtype {self.output_weights.dtype} does not match"
                f" the specified precision {self.config.precision}",
            )
        input_vocab_size, input_model_dim = self.input_weights.shape
        output_vocab_size, output_model_dim = self.output_weights.shape
        if input_vocab_size != output_vocab_size:
            raise ValueError(
                f"Input vocab size {input_vocab_size} does not match the output vocab size {output_vocab_size}",
            )
        if input_model_dim != output_model_dim:
            raise ValueError(
                f"Input model dim {input_model_dim} does not match the output model dim {output_model_dim}",
            )

    def _prepare_input_weights(self) -> Float[Array, "vocabulary channels"]:
        return self.input_weights

    def _prepare_output_weights(self) -> Float[Array, "vocabulary channels"]:
        return self.output_weights

    def export_weights(self) -> ParameterTree:
        return {
            "input_weights": self.input_weights,
            "output_weights": self.output_weights,
        }

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self:
        assert isinstance(weights, Mapping)
        return replace(
            self,
            input_weights=weights["input_weights"],
            output_weights=weights["output_weights"],
        )


@dataclass(frozen=True)
class QuantizedTiedEmbeddingConfig(EmbeddingConfigBase):
    embedding_quantization_mode: QuantizationMode
    activation_quantization_mode: QuantizationMode | None
    activation_precision: DTypeLike

    def random_init(
        self,
        vocab_size: int,
        model_dim: int,
        *,
        key: PRNGKeyArray,
    ) -> "QuantizedTiedEmbedding":
        min_val, max_val = self.embedding_quantization_mode.range
        min_abs_val = min(abs(min_val), abs(max_val))
        scale = 1 / min_abs_val
        scales = scale * jnp.ones(vocab_size, dtype=self.activation_precision)
        weights = jax.random.normal(key, (vocab_size, model_dim), dtype=self.activation_precision)
        weights = quantize_weights(weights * min_abs_val, self.embedding_quantization_mode)
        return QuantizedTiedEmbedding(config=self, weights=weights, scales=scales)

    def empty(
        self,
        vocab_size: int,
        model_dim: int,
    ) -> "QuantizedTiedEmbedding":
        scales = dummy_array(vocab_size, dtype=self.activation_precision)
        weights = dummy_array((vocab_size, model_dim), dtype=self.activation_precision)
        return QuantizedTiedEmbedding(config=self, weights=weights, scales=scales)


class QuantizedTiedEmbedding(EmbeddingBase[QuantizedTiedEmbeddingConfig]):
    weights: Float[Array, "vocabulary channels"]
    scales: Float[Array, " vocabulary"]

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.activation_precision

    @property
    def model_dim(self) -> int:
        _, model_dim = self.weights.shape
        return model_dim

    @property
    def vocab_size(self) -> int:
        vocab_size, _ = self.weights.shape
        return vocab_size

    def __post_init__(self) -> None:
        if self.weights.dtype != self.config.activation_precision:
            raise ValueError(
                f"Embedding dtype ({self.scales.dtype}) is not equal to specified activation precision"
                f" ({self.config.activation_precision})."
                " Quantized layers require parameter dtypes to be equal to the activation precision.",
            )
        if self.scales.dtype != self.config.activation_precision:
            raise ValueError(
                f"Scales dtype {self.scales.dtype} does not match the specified activation precision"
                f" {self.config.activation_precision}"
                " Quantized layers require parameter dtypes to be equal to the activation precision.",
            )
        weights_vocab_size, _ = self.weights.shape
        (scales_vocab_size,) = self.scales.shape
        if weights_vocab_size != scales_vocab_size:
            raise ValueError(
                f"Embedding vocab size {weights_vocab_size} does not match"
                f" the scales dimension size {scales_vocab_size}",
            )

    @property
    def int_weights(self) -> Int[Array, "vocabulary channels"]:
        quantized = quantize_weights(self.weights, self.config.embedding_quantization_mode)
        casted = quantized.astype(self.config.embedding_quantization_mode.dtype)

        if self.config.embedding_quantization_mode == QuantizationMode.UINT4:
            packed = jax_uint4_to_packed_uint8(casted)
        else:
            packed = casted

        return packed

    def _prepare_weights(self) -> Float[Array, "vocabulary channels"]:
        quantized_weights = quantize_weights(self.weights, self.config.embedding_quantization_mode)
        quantized_weights = quantized_weights * self.scales.reshape(-1, 1)
        return quantized_weights

    def _prepare_input_weights(self) -> Float[Array, "vocabulary channels"]:
        return self._prepare_weights()

    def _prepare_output_weights(self) -> Float[Array, "vocabulary channels"]:
        return self._prepare_weights()

    @eqx.filter_jit
    def readout(self, x: Float[Array, " channels"]) -> Float[Array, " vocabulary"]:
        if self.config.activation_quantization_mode is not None:
            x = dynamically_quantize_activations(x, self.config.activation_quantization_mode)
        return super().readout(x)

    def export_weights(self) -> ParameterTree:
        return {
            "weights": self.int_weights,
            "scales": self.scales,
        }

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self:
        assert isinstance(weights, Mapping)
        assert isinstance(weights["weights"], Array)
        stored_weights = weights["weights"]

        if self.config.embedding_quantization_mode == QuantizationMode.UINT4:
            stored_weights = jax_uint8_to_unpacked_uint4(stored_weights)

        return replace(
            self,
            weights=stored_weights.astype(self.weights.dtype),
            scales=weights["scales"],
        )


@dataclass(frozen=True)
class MLXQuantizedTiedEmbeddingConfig(EmbeddingConfigBase):
    group_size: int
    embedding_quantization_mode: QuantizationMode
    activation_quantization_mode: QuantizationMode | None
    activation_precision: DTypeLike

    def random_init(
        self,
        vocab_size: int,
        model_dim: int,
        *,
        key: PRNGKeyArray,
    ) -> "QuantizedTiedEmbedding":
        raise NotImplementedError

    def empty(
        self,
        vocab_size: int,
        model_dim: int,
    ) -> "MLXQuantizedTiedEmbedding":
        assert model_dim % self.group_size == 0
        model_groups = model_dim // self.group_size
        weights = dummy_array((vocab_size, model_dim), dtype=self.activation_precision)
        scales = dummy_array((vocab_size, model_groups), dtype=self.activation_precision)
        biases = dummy_array((vocab_size, model_groups), dtype=self.activation_precision)
        return MLXQuantizedTiedEmbedding(config=self, weights=weights, scales=scales, biases=biases)


class MLXQuantizedTiedEmbedding(EmbeddingBase[MLXQuantizedTiedEmbeddingConfig]):
    weights: Float[Array, "vocabulary channels"]
    scales: Float[Array, "vocabulary groups"]
    biases: Float[Array, "vocabulary groups"]

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.activation_precision

    @property
    def model_dim(self) -> int:
        _, model_dim = self.weights.shape
        return model_dim

    @property
    def vocab_size(self) -> int:
        vocab_size, _ = self.weights.shape
        return vocab_size

    @property
    def int_weights(self) -> Int[Array, "vocabulary channels"]:
        quantized = quantize_weights(self.weights, self.config.embedding_quantization_mode)
        casted = quantized.astype(self.config.embedding_quantization_mode.dtype)

        if self.config.embedding_quantization_mode == QuantizationMode.UINT4:
            packed = jax_uint4_to_packed_uint8(casted)
        else:
            packed = casted

        return packed

    def _prepare_weights(self) -> Float[Array, "vocabulary channels"]:
        quantized_weights = quantize_weights(self.weights, self.config.embedding_quantization_mode)
        grouped_weights = rearrange(
            quantized_weights,
            "vocab (groups elements) -> vocab groups elements",
            elements=self.config.group_size,
        )

        scales = rearrange(self.scales, "vocab groups -> vocab groups 1")

        biases = rearrange(self.biases, "vocab groups -> vocab groups 1")

        scaled_grouped_weights = grouped_weights * scales + biases

        result = rearrange(
            scaled_grouped_weights,
            "vocab groups elements -> vocab (groups elements)",
        )
        return result

    def _prepare_input_weights(self) -> Float[Array, "vocabulary channels"]:
        return self._prepare_weights()

    def _prepare_output_weights(self) -> Float[Array, "vocabulary channels"]:
        return self._prepare_weights()

    @eqx.filter_jit
    def readout(self, x: Float[Array, " channels"]) -> Float[Array, " vocabulary"]:
        if self.config.activation_quantization_mode is not None:
            x = dynamically_quantize_activations(x, self.config.activation_quantization_mode)
        return super().readout(x)

    def export_weights(self) -> ParameterTree:
        return {
            "weights": self.int_weights,
            "scales": self.scales,
            "biases": self.biases,
        }

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self:
        assert isinstance(weights, Mapping)
        assert isinstance(weights["weights"], Array)
        assert isinstance(weights["scales"], Array)
        assert isinstance(weights["biases"], Array)

        unpacked_weights = weights["weights"]

        if self.config.embedding_quantization_mode == QuantizationMode.UINT4:
            unpacked_weights = jax_uint8_to_unpacked_uint4(weights["weights"])

        return replace(
            self,
            weights=unpacked_weights.astype(self.weights.dtype),
            scales=weights["scales"],
            biases=weights["biases"],
        )


@dataclass(frozen=True)
class MLXQuantizedUntiedEmbeddingConfig(EmbeddingConfigBase):
    group_size: int
    embedding_quantization_mode: QuantizationMode
    activation_quantization_mode: QuantizationMode | None
    activation_precision: DTypeLike

    def random_init(
        self,
        vocab_size: int,
        model_dim: int,
        *,
        key: PRNGKeyArray,
    ) -> "MLXQuantizedUntiedEmbedding":
        raise NotImplementedError

    def empty(
        self,
        vocab_size: int,
        model_dim: int,
    ) -> "MLXQuantizedUntiedEmbedding":
        assert model_dim % self.group_size == 0
        model_groups = model_dim // self.group_size
        return MLXQuantizedUntiedEmbedding(
            config=self,
            input_weights=dummy_array((vocab_size, model_dim), dtype=self.activation_precision),
            input_scales=dummy_array((vocab_size, model_groups), dtype=self.activation_precision),
            input_biases=dummy_array((vocab_size, model_groups), dtype=self.activation_precision),
            output_weights=dummy_array((vocab_size, model_dim), dtype=self.activation_precision),
            output_scales=dummy_array((vocab_size, model_groups), dtype=self.activation_precision),
            output_biases=dummy_array((vocab_size, model_groups), dtype=self.activation_precision),
        )


class MLXQuantizedUntiedEmbedding(EmbeddingBase[MLXQuantizedUntiedEmbeddingConfig]):
    input_weights: Float[Array, "vocabulary channels"]
    input_scales: Float[Array, "vocabulary groups"]
    input_biases: Float[Array, "vocabulary groups"]
    output_weights: Float[Array, "vocabulary channels"]
    output_scales: Float[Array, "vocabulary groups"]
    output_biases: Float[Array, "vocabulary groups"]

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.activation_precision

    @property
    def model_dim(self) -> int:
        _, model_dim = self.input_weights.shape
        return model_dim

    @property
    def vocab_size(self) -> int:
        vocab_size, _ = self.input_weights.shape
        return vocab_size

    @property
    def int_input_weights(self) -> Int[Array, "vocabulary channels"]:
        quantized = quantize_weights(self.input_weights, self.config.embedding_quantization_mode)
        casted = quantized.astype(self.config.embedding_quantization_mode.dtype)

        if self.config.embedding_quantization_mode == QuantizationMode.UINT4:
            packed = jax_uint4_to_packed_uint8(casted)
        else:
            packed = casted

        return packed

    @property
    def int_output_weights(self) -> Int[Array, "vocabulary channels"]:
        quantized = quantize_weights(self.output_weights, self.config.embedding_quantization_mode)
        casted = quantized.astype(self.config.embedding_quantization_mode.dtype)

        if self.config.embedding_quantization_mode == QuantizationMode.UINT4:
            packed = jax_uint4_to_packed_uint8(casted)
        else:
            packed = casted

        return packed

    def _prepare_input_weights(self) -> Float[Array, "vocabulary channels"]:
        quantized_weights = quantize_weights(self.input_weights, self.config.embedding_quantization_mode)
        grouped_weights = rearrange(
            quantized_weights,
            "vocab (groups elements) -> vocab groups elements",
            elements=self.config.group_size,
        )

        scales = rearrange(self.input_scales, "vocab groups -> vocab groups 1")

        biases = rearrange(self.input_biases, "vocab groups -> vocab groups 1")

        scaled_grouped_weights = grouped_weights * scales + biases

        result = rearrange(
            scaled_grouped_weights,
            "vocab groups elements -> vocab (groups elements)",
        )
        return result

    def _prepare_output_weights(self) -> Float[Array, "vocabulary channels"]:
        quantized_weights = quantize_weights(self.output_weights, self.config.embedding_quantization_mode)
        grouped_weights = rearrange(
            quantized_weights,
            "vocab (groups elements) -> vocab groups elements",
            elements=self.config.group_size,
        )

        scales = rearrange(self.output_scales, "vocab groups -> vocab groups 1")

        biases = rearrange(self.output_biases, "vocab groups -> vocab groups 1")

        scaled_grouped_weights = grouped_weights * scales + biases

        result = rearrange(
            scaled_grouped_weights,
            "vocab groups elements -> vocab (groups elements)",
        )
        return result

    @eqx.filter_jit
    def readout(self, x: Float[Array, " channels"]) -> Float[Array, " vocabulary"]:
        if self.config.activation_quantization_mode is not None:
            x = dynamically_quantize_activations(x, self.config.activation_quantization_mode)
        return super().readout(x)

    def export_weights(self) -> ParameterTree:
        return {
            "input_weights": self.int_input_weights,
            "input_scales": self.input_scales,
            "input_biases": self.input_biases,
            "output_weights": self.int_output_weights,
            "output_scales": self.output_scales,
            "output_biases": self.output_biases,
        }

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self:
        assert isinstance(weights, Mapping)
        assert isinstance(weights["input_weights"], Array)
        assert isinstance(weights["input_scales"], Array)
        assert isinstance(weights["input_biases"], Array)
        assert isinstance(weights["output_weights"], Array)
        assert isinstance(weights["output_scales"], Array)
        assert isinstance(weights["output_biases"], Array)

        unpacked_input_weights = weights["input_weights"]
        unpacked_output_weights = weights["output_weights"]

        if self.config.embedding_quantization_mode == QuantizationMode.UINT4:
            unpacked_input_weights = jax_uint8_to_unpacked_uint4(weights["input_weights"])
            unpacked_output_weights = jax_uint8_to_unpacked_uint4(weights["output_weights"])

        return replace(
            self,
            input_weights=unpacked_input_weights.astype(self.input_weights.dtype),
            input_scales=weights["input_scales"],
            input_biases=weights["input_biases"],
            output_weights=unpacked_output_weights.astype(self.output_weights.dtype),
            output_scales=weights["output_scales"],
            output_biases=weights["output_biases"],
        )


@dataclass(frozen=True)
class MLXSemiQuantizedUntiedEmbeddingConfig(EmbeddingConfigBase):
    group_size: int
    embedding_quantization_mode: QuantizationMode
    activation_quantization_mode: QuantizationMode | None
    activation_precision: DTypeLike

    def random_init(
        self,
        vocab_size: int,
        model_dim: int,
        *,
        key: PRNGKeyArray,
    ) -> "MLXSemiQuantizedUntiedEmbedding":
        raise NotImplementedError

    def empty(
        self,
        vocab_size: int,
        model_dim: int,
    ) -> "MLXSemiQuantizedUntiedEmbedding":
        assert model_dim % self.group_size == 0
        model_groups = model_dim // self.group_size
        input_weights = dummy_array((vocab_size, model_dim), dtype=self.activation_precision)
        output_weights = dummy_array((vocab_size, model_dim), dtype=self.activation_precision)
        output_scales = dummy_array((vocab_size, model_groups), dtype=self.activation_precision)
        output_biases = dummy_array((vocab_size, model_groups), dtype=self.activation_precision)
        return MLXSemiQuantizedUntiedEmbedding(
            config=self,
            input_weights=input_weights,
            output_weights=output_weights,
            output_scales=output_scales,
            output_biases=output_biases,
        )


class MLXSemiQuantizedUntiedEmbedding(EmbeddingBase[MLXSemiQuantizedUntiedEmbeddingConfig]):
    input_weights: Float[Array, "vocabulary channels"]
    output_weights: Float[Array, "vocabulary channels"]
    output_scales: Float[Array, "vocabulary groups"]
    output_biases: Float[Array, "vocabulary groups"]

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.activation_precision

    @property
    def model_dim(self) -> int:
        _, model_dim = self.input_weights.shape
        return model_dim

    @property
    def vocab_size(self) -> int:
        vocab_size, _ = self.input_weights.shape
        return vocab_size

    @property
    def int_output_weights(self) -> Int[Array, "vocabulary channels"]:
        quantized = quantize_weights(self.output_weights, self.config.embedding_quantization_mode)
        casted = quantized.astype(self.config.embedding_quantization_mode.dtype)

        if self.config.embedding_quantization_mode == QuantizationMode.UINT4:
            packed = jax_uint4_to_packed_uint8(casted)
        else:
            packed = casted

        return packed

    def _prepare_input_weights(self) -> Float[Array, "vocabulary channels"]:
        return self.input_weights

    def _prepare_output_weights(self) -> Float[Array, "vocabulary channels"]:
        quantized_weights = quantize_weights(self.output_weights, self.config.embedding_quantization_mode)
        grouped_weights = rearrange(
            quantized_weights,
            "vocab (groups elements) -> vocab groups elements",
            elements=self.config.group_size,
        )

        scales = rearrange(self.output_scales, "vocab groups -> vocab groups 1")

        biases = rearrange(self.output_biases, "vocab groups -> vocab groups 1")

        scaled_grouped_weights = grouped_weights * scales + biases

        result = rearrange(
            scaled_grouped_weights,
            "vocab groups elements -> vocab (groups elements)",
        )
        return result

    @eqx.filter_jit
    def readout(self, x: Float[Array, " channels"]) -> Float[Array, " vocabulary"]:
        if self.config.activation_quantization_mode is not None:
            x = dynamically_quantize_activations(x, self.config.activation_quantization_mode)
        return super().readout(x)

    def export_weights(self) -> ParameterTree:
        return {
            "input_weights": self.input_weights,
            "output_weights": self.int_output_weights,
            "output_scales": self.output_scales,
            "output_biases": self.output_biases,
        }

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self:
        assert isinstance(weights, Mapping)
        assert isinstance(weights["input_weights"], Array)
        assert isinstance(weights["output_weights"], Array)
        assert isinstance(weights["output_scales"], Array)
        assert isinstance(weights["output_biases"], Array)

        unpacked_output_weights = weights["output_weights"]

        if self.config.embedding_quantization_mode == QuantizationMode.UINT4:
            unpacked_output_weights = jax_uint8_to_unpacked_uint4(weights["output_weights"])

        return replace(
            self,
            input_weights=weights["input_weights"],
            output_weights=unpacked_output_weights.astype(self.output_weights.dtype),
            output_scales=weights["output_scales"],
            output_biases=weights["output_biases"],
        )


EmbeddingConfig = (
    TiedEmbeddingConfig
    | UntiedEmbeddingConfig
    | QuantizedTiedEmbeddingConfig
    | MLXQuantizedTiedEmbeddingConfig
    | MLXQuantizedUntiedEmbeddingConfig
    | MLXSemiQuantizedUntiedEmbeddingConfig
)


register_config_union(EmbeddingConfig)  # type: ignore (pyright bug)



from collections.abc import Mapping
from dataclasses import dataclass
from typing import Literal

from jaxtyping import DTypeLike

from lalamo.modules import (
    AttentionConfig,
    DecoderConfig,
    DenseMLPConfig,
    FullPrecisionLinearConfig,
    GroupQuantizedLinearConfig,
    LlamaRoPEConfig,
    MLXQuantizedLinearConfig,
    MLXQuantizedTiedEmbeddingConfig,
    NormalizationConfig,
    SiLU,
    TiedEmbeddingConfig,
    TransformerConfig,
    TransformerLayerConfig,
    UnscaledRoPEConfig,
    UntiedEmbeddingConfig,
    UpcastMode,
    YARNRoPEConfig,
)
from lalamo.quantization import QuantizationMode

from .common import HuggingFaceLMConfig, MLXQuantizationConfig, QuantizationConfigType

__all__ = ["HFLlamaConfig"]


@dataclass(frozen=True)
class LlamaRopeScalingConfig:
    factor: float
    high_freq_factor: float
    low_freq_factor: float
    original_max_position_embeddings: int
    rope_type: Literal["llama3"]


@dataclass(frozen=True)
class YarnRopeScalingConfig:
    factor: float
    beta_fast: float
    beta_slow: float
    original_max_position_embeddings: int
    rope_type: Literal["yarn"]
    truncate: bool


@dataclass(frozen=True)
class HFLlamaConfig(HuggingFaceLMConfig):
    torch_dtype: Literal["bfloat16", "float16", "float32"]
    architectures: list[Literal["LlamaForCausalLM"]]
    attention_bias: bool
    attention_dropout: float
    bos_token_id: int | list[int]
    eos_token_id: int | list[int]
    hidden_act: Literal["silu"]
    hidden_size: int
    initializer_range: float
    intermediate_size: int
    max_position_embeddings: int
    mlp_bias: bool
    model_type: Literal["llama"]
    num_attention_heads: int
    num_hidden_layers: int
    num_key_value_heads: int
    pretraining_tp: int
    rms_norm_eps: float
    rope_scaling: LlamaRopeScalingConfig | YarnRopeScalingConfig | None
    rope_theta: float
    tie_word_embeddings: bool
    transformers_version: str
    use_cache: bool
    vocab_size: int
    head_dim: int | None = None

    quantization: QuantizationConfigType = None
    quantization_config: QuantizationConfigType = None

    def to_decoder_config(
        self,
        context_length: int | None,
        activation_precision: DTypeLike,
        accumulation_precision: DTypeLike,
        metadata_dict: Mapping[str, str],  # noqa: ARG002
    ) -> DecoderConfig:
        quantization = self.quantization or self.quantization_config
        if isinstance(quantization, MLXQuantizationConfig):
            assert self.tie_word_embeddings, "only tied embeddings are supported"
            embedding_config = MLXQuantizedTiedEmbeddingConfig(
                input_scale=None,
                logit_soft_cap=None,
                group_size=quantization.group_size,
                embedding_quantization_mode=QuantizationMode.from_num_bits(quantization.bits),
                activation_quantization_mode=None,
                activation_precision=activation_precision,
            )
        elif self.tie_word_embeddings:
            embedding_config = TiedEmbeddingConfig(
                input_scale=None,
                logit_soft_cap=None,
                precision=activation_precision,
            )
        else:
            embedding_config = UntiedEmbeddingConfig(
                input_scale=None,
                logit_soft_cap=None,
                precision=activation_precision,
            )
        if self.rope_scaling is None:
            rope_config = UnscaledRoPEConfig(
                precision=activation_precision,
                base=self.rope_theta,
                max_sequence_length=context_length or self.max_position_embeddings,
            )
        elif isinstance(self.rope_scaling, YarnRopeScalingConfig):
            rope_config = YARNRoPEConfig(
                precision=activation_precision,
                base=self.rope_theta,
                max_sequence_length=context_length or self.max_position_embeddings,
                scaling_factor=self.rope_scaling.factor,
                original_context_length=self.rope_scaling.original_max_position_embeddings,
                beta_fast=self.rope_scaling.beta_fast,
                beta_slow=self.rope_scaling.beta_slow,
                truncate=self.rope_scaling.truncate,
            )
        elif isinstance(self.rope_scaling, LlamaRopeScalingConfig):
            rope_config = LlamaRoPEConfig(
                precision=activation_precision,
                base=self.rope_theta,
                max_sequence_length=context_length or self.max_position_embeddings,
                scaling_factor=self.rope_scaling.factor,
                original_context_length=self.rope_scaling.original_max_position_embeddings,
                low_frequency_factor=self.rope_scaling.low_freq_factor,
                high_frequency_factor=self.rope_scaling.high_freq_factor,
            )
        else:
            raise ValueError("Unsupported rope_scaling configuration")
        rmsnorm_config = NormalizationConfig(
            scale_precision=activation_precision,
            accumulation_precision=accumulation_precision,
            epsilon=self.rms_norm_eps,
            scale_offset=None,
            upcast_mode=UpcastMode.ONLY_NORMALIZATION,
            subtract_mean=False,
        )
        if quantization is None:
            linear_config = FullPrecisionLinearConfig(
                precision=activation_precision,
            )
        elif isinstance(quantization, MLXQuantizationConfig):
            linear_config = MLXQuantizedLinearConfig(
                group_size=quantization.group_size,
                weight_quantization_mode=QuantizationMode.from_num_bits(quantization.bits),
                activation_quantization_mode=None,
                activation_precision=activation_precision,
            )
        else:
            linear_config = GroupQuantizedLinearConfig(
                group_size=quantization.group_size,
                weight_quantization_mode=QuantizationMode.from_num_bits(quantization.bits),
                activation_quantization_mode=None,
                activation_precision=activation_precision,
            )
        attention_config = AttentionConfig(
            qkv_projection_config=linear_config,
            out_projection_config=linear_config,
            query_norm_config=None,
            key_norm_config=None,
            logit_soft_cap=None,
            has_sinks=False,
            has_qkv_biases=self.attention_bias,
            has_out_biases=False,
            num_heads=self.num_attention_heads,
            num_groups=self.num_key_value_heads,
            head_dim=(self.head_dim if self.head_dim is not None else self.hidden_size // self.num_attention_heads),
            is_causal=True,
            scale=None,
            sliding_window_size=None,
        )
        mlp_config = DenseMLPConfig(
            linear_config=linear_config,
            activation=SiLU(),
            has_up_biases=False,
            has_down_biases=False,
            up_clipping=None,
            gate_clipping=None,
        )
        transformer_layer_config = TransformerLayerConfig(
            pre_mixer_norm_config=rmsnorm_config,
            mixer_config=attention_config,
            post_mixer_norm_config=None,
            pre_mlp_norm_config=rmsnorm_config,
            mlp_config=mlp_config,
            post_mlp_norm_config=None,
        )
        transformer_config = TransformerConfig(
            global_rope_config=rope_config,
            local_rope_config=None,
            layer_configs=(transformer_layer_config,) * self.num_hidden_layers,
            output_norm_config=rmsnorm_config,
            model_dim=self.hidden_size,
            hidden_dim=self.intermediate_size,
            context_length=context_length or self.max_position_embeddings,
        )
        return DecoderConfig(
            embedding_config=embedding_config,
            transformer_config=transformer_config,
            vocab_size=self.vocab_size,
        )
